"""This module implements tools that are generally useful in dantro"""

import os
import sys
import subprocess
import collections
import logging

import numpy as np

# Get a logger instance
log = logging.getLogger(__name__)

# Terminal, TTY-related
IS_A_TTY = sys.stdout.isatty()
try:
    _, TTY_COLS = subprocess.check_output(['stty', 'size']).split()
except:
    # Probably not run from terminal --> set value manually
    TTY_COLS = 79
else:
    TTY_COLS = int(TTY_COLS)
log.debug("Determined TTY_COLS: %d, IS_A_TTY: %d", TTY_COLS, IS_A_TTY)

# -----------------------------------------------------------------------------
# Import private yaml module, where everything is configured

from ._yaml import yaml, load_yml, write_yml


# -----------------------------------------------------------------------------
# Dictionary operations

def recursive_update(d: dict, u: dict) -> dict:
    """Recursively updates the Mapping-like object `d` with the Mapping-like
    object `u` and returns it. Note that this does not create a copy of `d`!
    
    Based on: http://stackoverflow.com/a/32357112/1827608
    
    Args:
        d (dict): The mapping to update
        u (dict): The mapping whose values are used to update `d`
    
    Returns:
        dict: The updated dict `d`
    """
    for k, v in u.items():
        if isinstance(d, collections.abc.Mapping):
            # Already a Mapping
            if isinstance(v, collections.abc.Mapping):
                # Already a Mapping, continue recursion
                d[k] = recursive_update(d.get(k, {}), v)
                # This already creates a mapping if the key was not available
            else:
                # Not a mapping -> at leaf -> update value
                d[k] = v    # ... which is just u[k]

        else:
            # Not a mapping -> create one
            d = {k: u[k]}
    return d


# -----------------------------------------------------------------------------
# Terminal messaging

def clear_line(only_in_tty=True, break_if_not_tty=True):
    """Clears the current terminal line and resets the cursor to the first
    position using a POSIX command.
    
    Based on: https://stackoverflow.com/a/25105111/1827608
    
    Args:
        only_in_tty (bool, optional): If True (default) will only clear the
            line if the script is executed in a TTY
        break_if_not_tty (bool, optional): If True (default), will insert a
            line break if the script is not executed in a TTY
    """
    # Differentiate cases
    if (only_in_tty and IS_A_TTY) or not only_in_tty:
        # Print the POSIX character
        print('\x1b[2K\r', end='')

    if break_if_not_tty and not IS_A_TTY:
        # print linebreak (no flush)
        print('\n', end='')

    # flush manually (there might not have been a linebreak)
    sys.stdout.flush()

def fill_line(s: str, *, num_cols: int=TTY_COLS, fill_char: str=" ",
              align: str="left") -> str:
    """Extends the given string such that it fills a whole line of `num_cols`
    columns.
    
    Args:
        s (str): The string to extend to a whole line
        num_cols (int, optional): The number of colums of the line; defaults to
            the number of TTY columns or – if those are not available – 79
        fill_char (str, optional): The fill character
        align (str, optional): The alignment. Can be: 'left', 'right', 'center'
            or the one-letter equivalents.
    
    Returns:
        str: The string of length `num_cols`
    
    Raises:
        ValueError: For invalid `align` or `fill_char` argument
    """
    if len(fill_char) != 1:
        raise ValueError("Argument `fill_char` needs to be string of length 1 "
                         "but was: "+str(fill_char))

    fill_str = fill_char * (num_cols - len(s))

    if align in ["left", "l", None]:
        return s + fill_str

    elif align in ["right", "r"]:
        return fill_str + s

    elif align in ["center", "centre", "c"]:
        return fill_str[:len(fill_str)//2] + s + fill_str[len(fill_str)//2:]

    raise ValueError("align argument '{}' not supported".format(align))

def center_in_line(s: str, *, num_cols: int=TTY_COLS, fill_char: str="·",
                   spacing: int=1) -> str:
    """Shortcut for a common fill_line use case.
    
    Args:
        s (str): The string to center in the line
        num_cols (int, optional): The number of columns in the line
        fill_char (str, optional): The fill character
        spacing (int, optional): The spacing around the string `s`
    
    Returns:
        str: The string centered in the line
    """
    spacing = " " * spacing
    return fill_line(spacing + s + spacing, num_cols=num_cols,
                     fill_char=fill_char, align='centre')


# -----------------------------------------------------------------------------
# Numerics

def apply_along_axis(func, axis: int, arr: np.ndarray,
                     *args, **kwargs) -> np.ndarray:
    """This is like numpy's function of the same name, but does not try to
    cast the results of func to an np.ndarray but tries to keep them as dtype
    object. Thus, the return value of this function will always have one fewer
    dimension then the input array.

    This goes along the equivalent formulation of np.apply_along_axis, outlined
    in their documentation of the function.
    """
    # Get the shapes of the outer and inner iteration; both are tuples!
    shape_outer, shape_inner = arr.shape[:axis], arr.shape[axis+1:]
    num_outer = len(shape_outer)

    # These together give the shape of the output array
    out = np.zeros(shape_outer + shape_inner, dtype='object')
    out.fill(None)

    log.debug("Applying function '%s' along axis ...", func.__name__)
    log.debug("  input array:     %s, %s", arr.shape, arr.dtype)
    log.debug("  axis to reduce:  %d", axis)
    log.debug("  output will be:  %s, %s", out.shape, out.dtype)

    # Now loop over the output array and at each position fill it with the
    # result of the function call.
    it = np.nditer(out, flags=('refs_ok', 'multi_index'))
    for _ in it:
        midx = it.multi_index

        # Build selector, which has the ellipsis at position `axis`, thus one
        # dimension higher than the out array and matching the input `arr`.
        sel = tuple(midx[:num_outer]) + (Ellipsis,) + tuple(midx[num_outer:])
        log.debug("  midx: %s  -->  selector: %s", midx, sel)

        # Apply function to selected parts of array, then write to the current
        # point in the iteration over the output array.
        out[midx] = func(arr[sel], *args, **kwargs)

    return out

# -----------------------------------------------------------------------------
# Fun with byte strings

def decode_bytestrings(obj) -> str:
    """Checks whether the given attribute value is or contains byte
    strings and if so, decodes it to a python string.
    
    Args:
        obj: The object to try to decode into holding python strings
    
    Returns:
        str: Either the unchanged object or the decoded one
    """
    # Check for data loaded as array of bytestring
    if isinstance(obj, np.ndarray):
        if obj.dtype.kind in ['S', 'a']:
            obj = obj.astype('U')

        # If it is of dtype object, decode all bytes objects
        if obj.dtype == np.dtype('object'):
            def decode_if_bytes(val):
                if isinstance(val, bytes):
                    return val.decode('utf8')
                return val

            # Apply element-wise
            obj = np.vectorize(decode_if_bytes)(obj)

    # ... or as bytes
    elif isinstance(obj, bytes):
        # Decode bytestring to unicode
        obj = obj.decode('utf8')

    return obj


# -----------------------------------------------------------------------------
# Misc

def is_iterable(obj) -> bool:
    """Tries whether the given object is iterable."""
    try:
        (e for e in obj)
    except:
        return False
    return True

def is_hashable(obj) -> bool:
    """Tries whether the given object is hashable."""
    try:
        hash(obj)
    except:
        return False
    return True

class DoNothingContext:
    """A context manager that ... does nothing."""

    def __init__(self):
        pass

    def __enter__(self):
        """Called upon entering the context using the ``with`` statement"""
        pass

    def __exit__(self, *args):
        """Called upon exiting context, with ``*args`` being exceptions etc."""
        pass




"""This module holds the abstract base classes needed for dantro"""

import abc
import collections
import collections.abc
from typing import Union, Tuple, Any
import logging

# Local constants
log = logging.getLogger(__name__)

# The character used for separating hierarchies in the path
PATH_JOIN_CHAR = "/"

# -----------------------------------------------------------------------------

class AbstractDataContainer(metaclass=abc.ABCMeta):
    """The AbstractDataContainer is the class defining the data container
    interface. It holds the bare basics of methods and attributes that _all_
    dantro data tree classes should have in common: a name, some data, and some
    association with others via an optional parent object.
    
    Via the parent and the name, path capabilities are provided. Thereby, each
    object in a data tree has some information about its location relative to
    a root object.
    Objects that have _no_ parent are regarded to be an object that is located
    "next to" root, i.e. having the path /<container_name>.
    """

    @abc.abstractmethod
    def __init__(self, *, name: str, data: Any):
        """Initialize the AbstractDataContainer, which holds the bare
        essentials of what a data container should have.
        
        Args:
            name (str): The name of this container
            data (Any): The data that is to be stored
        """
        self._logstr = None
        self._parent = None

        self._name = None
        self.name = name

        # Check data (raises if anything is off), then store it
        self._check_data(data)
        self._data = data

    # .........................................................................
    # Properties

    @property
    def name(self) -> str:
        """The name of this DataContainer-derived object."""
        return self._name

    @name.setter
    def name(self, new_name: str):
        """Rename this object; not always possible!"""
        # Check if this is a renaming operation; allow only if currently orphan
        if self._name is not None:
            if self.parent is not None:
                raise ValueError("Cannot rename {} to '{}', because a parent "
                                 "was already associated with it."
                                 "".format(self.logstr, new_name))

        # Require strings as name
        if not isinstance(new_name, str):
            raise TypeError("Name for {} needs to be a string, was of type "
                            "{} with value '{}'."
                            "".format(self.classname,
                                      type(new_name), new_name))

        # Ensure name does not contain path join character
        if PATH_JOIN_CHAR in new_name:
            raise ValueError("Name for {} cannot contain the path separator "
                             "'{}'! Got: '{}'"
                             "".format(self.classname, PATH_JOIN_CHAR,
                                       new_name))

        # Allow further checks by an additional method
        self._check_name(new_name)

        # Everything ok, store the attribute
        self._name = new_name

    @property
    def classname(self) -> str:
        """Returns the name of this DataContainer-derived class"""
        return self.__class__.__name__

    @property
    def logstr(self) -> str:
        """Returns the classname and name of this object; a combination often
        used in logging..."""
        # Store the cache value, if not already happened
        if self._logstr is None:
            self._logstr = "{} '{}'".format(self.classname, self.name)

        # Return the cache value
        return self._logstr

    @property
    def data(self):
        """The stored data."""
        return self._data

    @property
    def parent(self):
        """The associated parent of this container or group"""
        return self._parent

    @parent.setter
    def parent(self, cont):
        """Associate a new parent object with this container or group"""
        if self.parent is not None and cont is not None:
            raise ValueError("A parent was already associated with {cls:} "
                             "'{}'! Instead of manually setting the parent, "
                             "use the functions supplied to manipulate "
                             "members of this {cls:}."
                             "".format(self.name, cls=self.classname))
        
        log.trace("Setting %s as parent of %s ...",
                  cont.logstr if cont else None, self.logstr)
        self._parent = cont

    @property
    def path(self) -> str:
        """The path to get to this container or group from some root path"""
        if self.parent is None:
            # Is at the root, thus prefix it with the root character
            return PATH_JOIN_CHAR + self.name

        # else: not at the top, thus also need the parent's path
        return self.parent.path + PATH_JOIN_CHAR + self.name

    # .........................................................................
    # Item access

    @abc.abstractmethod
    def __getitem__(self, key):
        """Gets an item from the container."""

    @abc.abstractmethod
    def __setitem__(self, key, val) -> None:
        """Sets an item in the container."""

    @abc.abstractmethod
    def __delitem__(self, key) -> None:
        """Deletes an item from the container."""

    # .........................................................................
    # Helper functions

    def _check_name(self, new_name: str) -> None:
        """Called from name.setter and can be used to check the name that the
        container is supposed to have. On invalid name, this should raise.

        This method can be subclassed to implement more specific behaviour. To
        propagate the parent classes' behaviour the subclassed method should
        always call its parent method using super().
        
        Args:
            new_name (str): The new name, which is to be checked.
        """
        pass

    def _check_data(self, data: Any) -> None:
        """This method can be used to check the data provided to this container
        
        It is called before the data is stored in the __init__ method and
        should raise an exception or create a warning if the data is not as
        desired.

        This method can be subclassed to implement more specific behaviour. To
        propagate the parent classes' behaviour the subclassed method should
        always call its parent method using super().
        
        NOTE The CheckDataMixin provides a generalised implementation of this
             method to perform some type checks and react to unexpected types.
        
        Args:
            data (Any): The data to check
        """
        pass

    # .........................................................................
    # Formatting

    def __str__(self) -> str:
        """An info string, that describes the object. This invokes the
        formatting helpers to show the log string (type and name) as well as
        the info string of this object.
        """
        return "<{:logstr,info}>".format(self)

    def __repr__(self) -> str:
        """Same as __str__"""
        return str(self)

    def __format__(self, spec_str: str) -> str:
        """Creates a formatted string from the given specification.

        Invokes further methods which are prefixed by ``_format_``.
        """
        if not spec_str:
            return str(self)

        specs = spec_str.split(",")
        parts = []
        join_char = ", "

        for spec in specs:
            try:
                format_func = getattr(self, '_format_'+spec)
            except AttributeError as err:
                raise ValueError("No format string specification '{}', part "
                                 "of '{}', is available for {}!"
                                 "".format(spec, specs,
                                           self.logstr)) from err
            else:
                parts.append(format_func())

        return join_char.join(parts)

    def _format_name(self) -> str:
        """A __format__ helper function: returns the name"""
        return self.name
    
    def _format_cls_name(self) -> str:
        """A __format__ helper function: returns the class name"""
        return self.classname

    def _format_logstr(self) -> str:
        """A __format__ helper function: returns the log string, a combination
        of class name and name"""
        return self.logstr

    def _format_path(self) -> str:
        """A __format__ helper function: returns the path to this container"""
        return self.path

    @abc.abstractmethod
    def _format_info(self) -> str:
        """A __format__ helper function: returns an info string that is used
        to characterise this object. Should NOT include name and classname!
        """


# -----------------------------------------------------------------------------

class AbstractDataGroup(AbstractDataContainer, collections.abc.MutableMapping):
    """The AbstractDataGroup is the abstract basis of all data groups.

    It enforces a MutableMapping interface with a focus on _setting_ abilities
    and less so on deletion."""

    @property
    def data(self):
        """The stored data."""
        raise AttributeError("Cannot directly access group data!")

    @abc.abstractmethod
    def add(self, *conts, overwrite: bool=False) -> None:
        """Adds the given containers to the group."""

    @abc.abstractmethod
    def __contains__(self, cont: Union[str, AbstractDataContainer]) -> bool:
        """Whether the given container is a member of this group"""

    @abc.abstractmethod
    def keys(self):
        """Returns an iterator over the container names in this group."""

    @abc.abstractmethod
    def values(self):
        """Returns an iterator over the containers in this group."""

    @abc.abstractmethod
    def items(self):
        """Returns an iterator over the (name, data container) tuple of this
        group."""

    @abc.abstractmethod
    def get(self, key, default=None):
        """Return the container at `key`, or `default` if container with name
        `key` is not available."""

    @abc.abstractmethod
    def setdefault(self, key, default=None):
        """If `key` is in the dictionary, return its value. If not, insert
        `key` with a value of `default` and return `default`. `default`
        defaults to None."""

    @abc.abstractmethod
    def recursive_update(self, other):
        """Updates the group with the contents of another group."""

    @abc.abstractmethod
    def _format_tree(self) -> str:
        """A __format__ helper function: tree representation of this group"""

    @abc.abstractmethod
    def _tree_repr(self, level: int=0) -> str:
        """Recursively creates a multi-line string tree representation of this
        group. This is used by, e.g., the _format_tree method."""

# -----------------------------------------------------------------------------

class AbstractDataAttrs(collections.abc.Mapping, AbstractDataContainer):
    """The BaseDataAttrs class defines the interface for the `.attrs`
    attribute of a data container.

    This class derives from the abstract class as otherwise there would be 
    circular inheritance. It stores the attributes as mapping and need not be 
    subclassed.
    """
    
    # .........................................................................
    # Specify the attrs interface, dict-like

    @abc.abstractmethod
    def __contains__(self, key) -> bool:
        """Whether the given key is contained in the attributes."""
    
    @abc.abstractmethod
    def __len__(self) -> int:
        """The number of attributes."""

    @abc.abstractmethod
    def keys(self):
        """Returns an iterator over the attribute names."""

    @abc.abstractmethod
    def values(self):
        """Returns an iterator over the attribute values."""

    @abc.abstractmethod
    def items(self):
        """Returns an iterator over the (keys, values) tuple of the attributes.
        """

# -----------------------------------------------------------------------------

class AbstractDataProxy(metaclass=abc.ABCMeta):
    """A data proxy fills in for the place of a data container, e.g. if data
    should only be loaded on demand. It needs to supply the resolve method.
    """

    @abc.abstractmethod
    def __init__(self, obj):
        """Initialize the proxy object, being supplied with the object that
        this proxy is to be proxy for.
        """

    @property
    def classname(self) -> str:
        """Returns this proxy's class name"""
        return self.__class__.__name__

    @abc.abstractmethod
    def resolve(self, *, astype: type=None):
        """Get the data that this proxy is a placeholder for and return it.

        Note that this method does not place the resolved data in the
        container of which this proxy object is a placeholder for! This only
        returns the data.
        """

    @property
    @abc.abstractmethod
    def tags(self) -> Tuple[str]:
        """The tags describing this proxy object"""
    


# -----------------------------------------------------------------------------

class AbstractPlotCreator(metaclass=abc.ABCMeta):
    """This class defines the interface for PlotCreator classes"""

    @abc.abstractmethod
    def __init__(self, name: str, *, dm, **plot_cfg):
        """Initialize the PlotCreator, given a data manager, the plot name, and
        the default plot configuration.
        """

    @abc.abstractmethod
    def __call__(self, *, out_path: str=None, **update_plot_cfg):
        """Perform the plot, updating the configuration passed to __init__
        with the given values and then calling _plot.
        """

    @abc.abstractmethod
    def plot(self, out_path: str=None, **cfg):
        """Given a specific configuration, perform a plot.

        This method should always be private and only be called from __call__.
        """

    @abc.abstractmethod
    def get_ext(self) -> str:
        """Returns the extension to use for the upcoming plot"""

    @abc.abstractmethod
    def prepare_cfg(self, *, plot_cfg: dict, pspace) -> tuple:
        """Prepares the plot configuration for the PlotManager.

        This function is called by the plot manager before the first plot
        is created.
        
        The base implementation just passes the given arguments through.
        However, it can be re-implemented by derived classes to change the
        behaviour of the plot manager, e.g. by converting a plot configuration
        to a parameter space.
        """
        
    @abc.abstractmethod
    def can_plot(self, creator_name: str, **plot_cfg) -> bool:
        """Whether this plot creator is able to make a plot for the given plot
        configuration.

        This function is used by the PlotManager's auto-detect feature.
        
        Args:
            creator_name (str): The name for this creator used within the
                PlotManager.
            **plot_cfg: The plot configuration with which to decide this.
        
        Returns:
            bool: Whether this creator can be used for the given plot config
        """

    @abc.abstractmethod
    def _prepare_path(self, out_path: str) -> str:
        """Prepares the output path, creating directories if needed, then
        returning the full absolute path.

        This is called from __call__ and is meant to postpone directory
        creation as far as possible.
        """



"""Implements loading of Hdf5 files into the dantro data tree"""

import logging

import numpy as np
import h5py as h5

from ..base import BaseDataGroup, BaseDataContainer
from ..containers import NumpyDataContainer
from ..groups import OrderedDataGroup
from ..proxy import Hdf5DataProxy
from ..tools import fill_line, decode_bytestrings
from ._tools import add_loader

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------

class Hdf5LoaderMixin:
    """Supplies functionality to load hdf5 files into the data manager.
    
    It resolves the hdf5 groups into corresponding data groups and the datasets
    into NumpyDataContainers.

    If ``enable_mapping`` is set, the class variables ``_HDF5_DSET_MAP`` and
    ``_HDF5_GROUP_MAP`` are used to map from a string to a container type. The
    class variable ``_HDF5_MAP_FROM_ATTR`` determines the default value of the
    attribute to read and use as input string for the mapping. 
    """

    # Define the container classes to use when loading data with this mixin
    # The default class to use for containers
    _HDF5_DSET_DEFAULT_CLS = NumpyDataContainer

    # The mapping of types to data groups
    _HDF5_GROUP_MAP = None

    # The mapping of types to data containers
    _HDF5_DSET_MAP = None

    # The name of the attribute to read for mapping
    _HDF5_MAP_FROM_ATTR = None

    # Whether to decode strings stored as byte arrays to regular python strings
    _HDF5_DECODE_ATTR_BYTESTRINGS = True


    @add_loader(TargetCls=OrderedDataGroup, omit_self=False)
    def _load_hdf5(self, filepath: str, *,
                   TargetCls: type,
                   load_as_proxy: bool=False,
                   proxy_kwargs: dict=None,
                   lower_case_keys: bool=False,
                   enable_mapping: bool=False, 
                   map_from_attr: str=None,
                   print_params: dict=None
                   ) -> OrderedDataGroup:
        """Loads the specified hdf5 file into DataGroup- and DataContainer-like
        objects; this completely recreates the hierarchic structure of the hdf5
        file. The data can be loaded into memory completely, or be loaded as
        a proxy object.
        
        The h5py File and Group objects will be converted to the specified
        DataGroup-derived objects; the Dataset objects to the specified
        DataContainer-derived object.
        
        All HDF5 group or dataset attributes are carried over and are
        accessible under the ``attrs`` attribute of the respective dantro
        objects in the tree.
        
        Args:
            filepath (str): The path to the HDF5 file that is to be loaded
            TargetCls (type): The group type this is loaded into
            load_as_proxy (bool, optional): if True, the leaf datasets are
                loaded as :py:class:`dantro.proxy.hdf5.Hdf5DataProxy`
                objects. That way, the data is only loaded into memory when
                their ``.data`` property is accessed the first time, either
                directly or indirectly.
            proxy_kwargs (dict, optional): When loading as proxy, these
                parameters are unpacked in the ``__init__`` call. For available
                argument see :py:class:`dantro.proxy.hdf5.Hdf5DataProxy`.
            lower_case_keys (bool, optional): whether to use only lower-case
                versions of the paths encountered in the HDF5 file.
            enable_mapping (bool, optional): If true, will use the class
                variables _HDF5_GROUP_MAP and _HDF5_DSET_MAP to map groups or
                datasets to a custom container class during loading. Which
                attribute to read is determined by the `map_from_attr` argument
            map_from_attr (str, optional): From which attribute to read the
                key that is used in the mapping. If nothing is given, the
                class variable _HDF5_MAP_FROM_ATTR is used.
            print_params (dict, optional): parameters for the status report.
                Available keys:

                level (int):
                    how verbose to print loading info; possible values are:
                    ``0``: None, ``1``: on file level, ``2``: on dataset level
                fstr1:
                    format string level 1, receives keys ``name`` and ``file``,
                    which is the file path.
                fstr2:
                    format string level 2, receives keys ``name``, ``file`` and
                    ``obj``, which is an ``h5py.Dataset``.
        
        Returns:
            OrderedDataGroup: The populated root-level group, corresponding to
                the base group of the file
        
        Raises:
            ValueError: If `enable_mapping`, but no map attribute can be
                determined from the given argument or the class variable
                _HDF5_MAP_FROM_ATTR
        """
        
        def recursively_load_hdf5(src: h5.Group, target: BaseDataGroup, *,
                                  load_as_proxy: bool,
                                  proxy_kwargs: dict, 
                                  lower_case_keys: bool,
                                  DsetCls: BaseDataContainer,
                                  enable_mapping: bool, GroupMap: dict,
                                  DsetMap: dict,map_attr: str):
            """Recursively loads the data from the source hdf5 file into the 
            target DataGroup object.
            If given, each group or dataset is checked whether an attribute
            `container_type` or `dset_type` exists, which is then used to
            apply a mapping from that attribute to a certain type of DataGroup
            or DataContainer, respectively.
            
            Args:
                src (h5.Group): The source group to iterate over
                target (BaseDataGroup): The target group where the content from
                    the source group is loaded into
                load_as_proxy (bool): Whether to load as Hdf5DataProxy
                proxy_kwargs (dict): Unpacked in Hdf5DataProxy.__init__
                lower_case_keys (bool): Whether to make keys lower-case
                DsetCls (BaseDataContainer): The type that is used to create
                    the dataset-equivalents in ``target``
                enable_mapping (bool): Whether type mapping should be used
                GroupMap (dict): Map of names to BaseDataGroup-derived types
                DsetMap (dict): Map of names to BaseDataContainer-derived types
                map_attr (str): The HDF5 attribute to inspect in order to
                    determine the name of the mapping
            
            Raises:
                NotImplementedError: When encountering objects other than
                    groups or datasets in the HDF5 file
            """

            def get_map_attr_val(attrs) -> str:
                """Make sure the map attribute isn't a 1-sized array!"""
                attr_val = attrs[map_attr] # map_attr and check in outer scope

                if isinstance(attr_val, np.ndarray):
                    # Need be single item and already decoded
                    attr_val = attr_val.item()

                return attr_val

            def decode_attr_val(attr_val) -> str:
                """Wrapper around decode_bytestrings"""
                # If feature not activated, return without doing anything
                if not self._HDF5_DECODE_ATTR_BYTESTRINGS:
                    return attr_val

                return decode_bytestrings(attr_val)

            # Go through the elements of the source object
            for key, obj in src.items():
                if lower_case_keys and isinstance(key, str):
                    key = key.lower()

                if isinstance(obj, h5.Group):
                    # Need to continue recursion
                    # Extract attributes manually
                    attrs = {k: decode_attr_val(v)
                             for k, v in obj.attrs.items()}

                    # Determine the class to use for this group
                    if enable_mapping and GroupMap and attrs.get(map_attr):
                        # Try to resolve the mapping
                        try:
                            _GroupCls = GroupMap[get_map_attr_val(attrs)]

                        except KeyError:
                            # Fall back to default
                            log.warning("Could not find a mapping from map "
                                        "attribute %s='%s' (originally %s) "
                                        "to a DataGroup class. Available "
                                        "keys: %s. Falling back to default "
                                        "class ...",
                                        map_attr, get_map_attr_val(attrs),
                                        attrs[map_attr],
                                        ", ".join([k
                                                   for k in GroupMap.keys()]))
                            _GroupCls = None
                    
                    else:
                        # Use the default of the target group
                        _GroupCls = None
                        # None value will lead to new_group determining the
                        # type of this group. Unlike for datasets, this is
                        # always possible, because the type of the target
                        # group is used as fallback type.

                    # Create and add the group, passing the attributes
                    target.new_group(path=key, Cls=_GroupCls,
                                     attrs=attrs)

                    # Continue recursion
                    recursively_load_hdf5(obj, target[key],
                                          load_as_proxy=load_as_proxy,
                                          proxy_kwargs=proxy_kwargs,
                                          lower_case_keys=lower_case_keys,
                                          DsetCls=DsetCls,
                                          enable_mapping=enable_mapping,
                                          GroupMap=GroupMap, DsetMap=DsetMap,
                                          map_attr=map_attr)

                elif isinstance(obj, h5.Dataset):
                    # Reached a leaf -> Import the data and attributes into a
                    # BaseDataContainer-derived object. This assumes that the
                    # given DsetCls supports np.ndarray-like data

                    # Progress information on loading this dataset
                    if plvl >= 2:
                        line = fstr2.format(name=target.name, key=key, obj=obj)
                        print(fill_line(line), end="\r")

                    if load_as_proxy:
                        # Instantiate a proxy object
                        data = Hdf5DataProxy(obj,
                                             **(proxy_kwargs if proxy_kwargs
                                                else {}))
                    else:
                        # Import the data completely
                        data = np.array(obj)

                    # Extract attributes manually
                    attrs = {k: decode_attr_val(v)
                             for k, v in obj.attrs.items()}

                    # Determine the class to use for this dataset
                    if enable_mapping and DsetMap and attrs.get(map_attr):
                        # Try to resolve the mapping
                        try:
                            _DsetCls = DsetMap[get_map_attr_val(attrs)]

                        except KeyError:
                            # Fall back to default
                            log.warning("Could not find a mapping from map "
                                        "attribute %s='%s' (originally %s) to "
                                        "a DataContainer class. Available "
                                        "keys: %s. "
                                        "Falling back to default class %s...",
                                        map_attr, get_map_attr_val(attrs),
                                        attrs[map_attr],
                                        ", ".join([k for k in DsetMap.keys()]),
                                        DsetCls.__name__)
                            _DsetCls = DsetCls
                    
                    else:
                        # If the target group supplies a default container
                        # type, use that one; otherwise fall back to default.
                        if target._NEW_CONTAINER_CLS is not None:
                            _DsetCls = None
                            # Leads to new_container taking care of the type

                        else:
                            _DsetCls = DsetCls

                    # Now create and add the dataset, passing data and attrs
                    target.new_container(path=key, Cls=_DsetCls,
                                         data=data, attrs=attrs)

                else:
                    raise NotImplementedError("Object {} is neither a dataset "
                                              "nor a group, but of type {}. "
                                              "Cannot load this!"
                                              "".format(key, type(obj)))

        # Prepare print format strings
        print_params = print_params if print_params else {}
        plvl = print_params.get('level', 0)
        fstr1 = print_params.get('fstr1', "  Loading {name:} ... ")
        fstr2 = print_params.get('fstr2', "  Loading {name:} - {key:} ...")

        # Initialize the root group
        log.debug("Loading hdf5 file %s into %s ...",
                  filepath, TargetCls.__name__)
        root = TargetCls()

        # Get the classes to use for groups and/or containers
        DsetCls = self._HDF5_DSET_DEFAULT_CLS

        # Determine from which attribute to read the mapping
        if not map_from_attr:
            # No custom value was given; use the class variable, if available
            if self._HDF5_MAP_FROM_ATTR:
                map_from_attr = self._HDF5_MAP_FROM_ATTR

            elif enable_mapping:
                # Mapping was enabled but it is unclear from which attribute
                # the map should be read. Need to raise an exception
                raise ValueError("Could not determine from which attribute "
                                 "to read the mapping. Either set the loader "
                                 "argument `map_from_attr`, the class "
                                 "variable _HDF5_MAP_FROM_ATTR, or disable "
                                 "mapping altogether via the `enable_mapping` "
                                 "argument.")

        # Now recursively go through the hdf5 file and add them to the roo
        with h5.File(filepath, 'r') as h5file:
            if plvl >= 1:
                # Print information on the level of this file
                line = fstr1.format(name=root.name, file=filepath)
                print(fill_line(line), end="\r")

            # Load the file level attributes, manually re-creating the dict
            root.attrs = {k:v for k, v in h5file.attrs.items()}

            # Now recursively load the data into the root group
            recursively_load_hdf5(h5file, root,
                                  load_as_proxy=load_as_proxy,
                                  proxy_kwargs=proxy_kwargs,
                                  lower_case_keys=lower_case_keys,
                                  DsetCls=DsetCls,
                                  enable_mapping=enable_mapping,
                                  GroupMap=self._HDF5_GROUP_MAP,
                                  DsetMap=self._HDF5_DSET_MAP,
                                  map_attr=map_from_attr)

        return root

    @add_loader(TargetCls=OrderedDataGroup, omit_self=False)
    def _load_hdf5_proxy(self, *args, **kwargs) -> OrderedDataGroup:
        """This is a shorthand for
        :py:meth:`~dantro.data_loaders.load_hdf5.Hdf5LoaderMixin._load_hdf5`
        with the ``load_as_proxy`` flag set.
        """
        return self._load_hdf5(*args, load_as_proxy=True, **kwargs)

    @add_loader(TargetCls=OrderedDataGroup, omit_self=False)
    def _load_hdf5_as_dask(self, *args, **kwargs) -> OrderedDataGroup:
        """This is a shorthand for
        :py:meth:`~dantro.data_loaders.load_hdf5.Hdf5LoaderMixin._load_hdf5`
        with the ``load_as_proxy`` flag set and ``resolve_as_dask`` passed as
        additional arguments to the proxy via ``proxy_kwargs``.
        """
        return self._load_hdf5(*args,
                               load_as_proxy=True,
                               proxy_kwargs=dict(resolve_as_dask=True),
                               **kwargs)



"""The containers sub-package implements BaseDataContainer specializations"""

from .general import ObjectContainer, PassthroughContainer
from .general import MutableSequenceContainer, MutableMappingContainer
from .general import StringContainer
from .link import LinkContainer

from .numeric import NumpyDataContainer
from .xrdatactr import XrDataContainer



"""This module implements data processing operations for dantro objects"""

import logging
import operator
from importlib import import_module as _import_module
from difflib import get_close_matches
from typing import Callable, Any, Sequence, Union

import numpy as np
import xarray as xr

from .ordereddict import KeyOrderedDict
from ..base import BaseDataContainer, BaseDataGroup
from ..tools import apply_along_axis

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Some general helper operations

# Define boolean operators separately (might be useful elsewhere) & register
BOOLEAN_OPERATORS = {
    '==': operator.eq,  'eq': operator.eq,
    '<':  operator.lt,  'lt': operator.lt,
    '<=': operator.le,  'le': operator.le,
    '>':  operator.gt,  'gt': operator.gt,
    '>=': operator.ge,  'ge': operator.ge,
    '!=': operator.ne,  'ne': operator.ne,
    '^':  operator.xor, 'xor': operator.xor,
    # Expecting an iterable as second argument
    'in':               (lambda x, y: x in y),
    'not in':           (lambda x, y: x not in y),
    # Performing bitwise boolean operations to support numpy logic
    'in interval':      (lambda x, y: x >= y[0] & x <= y[1]),
    'not in interval':  (lambda x, y: x < y[0] | x > y[1]),
} # End of boolean operator definitions

# .............................................................................

def print_data(data: Any) -> Any:
    """Prints and passes on the data.

    The print operation distinguishes between dantro types (in which case some
    more information is shown) and  non-dantro types.
    """
    # Distinguish between dantro types and others
    if isinstance(data, BaseDataContainer):
        print("{}, with data:\n{}\n".format(data, data.data))
    
    elif isinstance(data, BaseDataGroup):
        print("{}\n".format(data.tree))

    else:
        print(data)

    return data

def import_module_or_object(module: str=None, name: str=None):
    """Imports a module or an object using the specified module string and the
    object name.
    
    Args:
        module (str, optional): A module string, e.g. numpy.random. If this is
            not given, it will import from the :py:mod`builtins` module. Also,
            relative module strings are resolved from :py:mod:`dantro`.
        name (str, optional): The name of the object to retrieve from the
            chosen module and return. This may also be a dot-separated sequence
            of attribute names which can be used to traverse along attributes.
    
    Returns:
        The chosen module or object, i.e. the object found at <module>.<name>
    
    Raises:
        AttributeError: In cases where part of the ``name`` argument could not
            be resolved due to a bad attribute name.
    """
    module = module if module else 'builtins'
    mod = _import_module(module, package='dantro') if module else __builtin__

    if not name:
        return mod

    # Get the object by traversing along the attributes of the module. By
    # allowing a name to be a sequence, object imports become more versatile.
    # The first object to get the name from is the module itself:
    obj = mod

    for attr_name in name.split("."):
        try:
            obj = getattr(obj, attr_name)
        
        except AttributeError as err:
            raise AttributeError("Failed to retrieve attribute or attribute "
                                 "sequence '{}' from module '{}'! "
                                 "Intermediate {} {} has no attribute '{}'!"
                                 "".format(name, mod.__name__,
                                           type(obj).__name__, obj, attr_name)
                                 ) from err

    return obj


# .............................................................................
# numpy and xarray operations

def create_mask(data: xr.DataArray,
                operator_name: str,
                rhs_value: float) -> xr.DataArray:
    """Given the data, returns a binary mask by applying the following
    comparison: ``data <operator> rhs value``.
    
    Args:
        data (xr.DataArray): The data to apply the comparison to. This is the
            lhs of the comparison.
        operator_name (str): The name of the binary operator function as
            registered in the ``BOOLEAN_OPERATORS`` constant.
        rhs_value (float): The right-hand-side value
    
    Raises:
        KeyError: On invalid operator name
    
    Returns:
        xr.DataArray: Boolean mask
    """
    # Get the operator function
    try:
        comp_func = BOOLEAN_OPERATORS[operator_name]

    except KeyError as err:
        raise KeyError("No boolean operator '{}' available! "
                       "Available operators: {}"
                       "".format(operator_name,
                                 ", ".join(BOOLEAN_OPERATORS.keys()))
                       ) from err

    # Apply the comparison
    data = comp_func(data, rhs_value)

    # Create a new name
    name = data.name + " (masked by '{} {}')".format(operator_name, rhs_value)

    # Build a new xr.DataArray from that data, retaining all information
    return xr.DataArray(data=data,
                        name=name,
                        dims=data.dims,
                        coords=data.coords)

def where(data: xr.DataArray,
          operator_name: str,
          rhs_value: float) -> xr.DataArray:
    """Filter elements from the given data according to a condition. Only
    those elemens where the condition is fulfilled are not masked.

    NOTE This leads to a dtype change to float.
    """
    # Get the mask and apply it
    return data.where(create_mask(data,
                                  operator_name=operator_name,
                                  rhs_value=rhs_value))

def count_unique(data) -> xr.DataArray:
    """Applies np.unique to the given data and constructs a xr.DataArray for
    the results.
    """
    unique, counts = np.unique(data, return_counts=True)

    # Construct a new data array and return
    return xr.DataArray(data=counts,
                        name=data.name + " (unique counts)",
                        dims=('unique',),
                        coords=dict(unique=unique))


# .............................................................................
# Working with multidimensional data, mostly xarray-based

def populate_ndarray(*objs, shape: tuple, dtype: str='float', order: str='C'
                     ) -> np.ndarray:
    """Populates an empty np.ndarray of the given dtype with the objects.
    
    Args:
        *objs: The objects to add to the 
        shape (tuple): The shape of the new array
        dtype (str, optional): Data type of the new array
        order (str, optional): Order of the new array
    
    Returns:
        np.ndarray: The newly created and populated array
    
    Raises:
        ValueError: If the number of given objects did not match the array size
    """
    arr = np.empty(shape, dtype=dtype, order=order)

    if len(objs) != arr.size:
        raise ValueError("Mismatch between array size ({}, shape: {}) and "
                         "number of given objects ({})!"
                         "".format(arr.size, arr.shape, len(objs)))

    it = np.nditer(arr, flags=('multi_index', 'refs_ok'))
    for obj, _ in zip(objs, it):
        arr[it.multi_index] = obj

    return arr


def multi_concat(arrs: np.ndarray, *, dims: Sequence[str]) -> xr.DataArray:
    """Concatenates ``xr.Dataset`` or ``xr.DataArray`` objects using
    ``xr.concat``. This function expects the xarray objects to be pre-aligned
    inside the numpy *object* array ``arrs``, with the number of dimensions
    matching the number of concatenation operations desired.
    The position inside the array carries information on where the objects that
    are to be concatenated are placed inside the higher dimensional coordinate
    system.

    Through multiple concatenation, the dimensionality of the contained objects
    is increased by ``dims``, while their dtype can be maintained.

    For the sequential application of ``xr.concat`` along the outer dimensions,
    the custom :py:func:`dantro.tools.apply_along_axis` is used.
    
    Args:
        arrs (np.ndarray): The array containing xarray objects which are to be
            concatenated. Each array dimension should correspond to one of the
            given ``dims``. For each of the dimensions, the ``xr.concat``
            operation is applied along the axis, effectively reducing the
            dimensionality of ``arrs`` to a scalar and increasing the
            dimensionality of the contained xarray objects until they
            additionally contain the dimensions specified in ``dims``.
        dims (Sequence[str]): A sequence of dimension names that is assumed to
            match the dimension names of the array. During each concatenation
            operation, the name is passed along to ``xr.concat`` where it is
            used to select the dimension of the *content* of ``arrs`` along
            which concatenation should occur.
    
    Raises:
        ValueError: If number of dimension names does not match the number of
            data dimensions.
    """
    # Check dimensionality
    if len(dims) != arrs.ndim:
        raise ValueError("The given sequence of dimension names, {}, did not "
                         "match the number of dimensions of data of shape {}!"
                         "".format(dims, arrs.shape))

    # Reverse-iterate over dimensions and concatenate them
    for dim_idx, dim_name in reversed(list(enumerate(dims))):
        log.debug("Concatenating along axis '%s' (idx: %d) ...",
                  dim_name, dim_idx)

        arrs = apply_along_axis(xr.concat, axis=dim_idx, arr=arrs,
                                dim=dim_name)
        # NOTE ``np.apply_along_axis`` would be what is desired here, but that
        #      function unfortunately tries to cast objects to np.arrays which
        #      is not what we want here at all! Thus, this function uses the
        #      custom dantro function of the same name instead.

    # Should be scalar now, get the element.
    return arrs.item()

def merge(arrs: Union[Sequence[Union[xr.DataArray, xr.Dataset]], np.ndarray],
          *, reduce_to_array: bool=False, **merge_kwargs
          ) -> Union[xr.Dataset, xr.DataArray]:
    """Merges the given sequence of xarray objects into an xr.Dataset.

    As a convenience, this also allows passing a numpy object array containing
    the xarray objects. Furthermore, if the resulting Dataset contains only a
    single data variable, that variable can be extracted as a DataArray which
    is then the return value of this operation.
    """
    if isinstance(arrs, np.ndarray):
        arrs = arrs.flat

    dset = xr.merge(arrs, **merge_kwargs)

    if not reduce_to_array:
        return dset
    
    elif len(dset.data_vars) != 1:
        raise ValueError("Can only reduce the Dataset resulting from the "
                         "xr.merge operation to a DataArray if one and only "
                         "one data variable is present in the Dataset! "
                         "Got: {}. Full data:\n{}"
                         "".format(", ".join(dset.data_vars), dset))

    # Get the name of the single data variable and then get the DataArray
    darr = dset[list(dset.data_vars.keys())[0]]
    # NOTE This is something else than the Dataset.to_array() method, which
    #      includes the name of the data variable as another coordinate. This
    #      is not desired, because it is not relevant.
    return darr


def expand_dims(d: Any, *, dim: dict=None, **kwargs) -> xr.DataArray:
    """Expands the dimensions of the given object.

    If the object does not support the `expand_dims` method, it will be
    attempted to convert it to an xr.DataArray.
    """
    if not hasattr(d, 'expand_dims'):
        d = xr.DataArray(d)

    return d.expand_dims(dim, **kwargs)


# -----------------------------------------------------------------------------
# The Operations Database -----------------------------------------------------
# NOTE If a single "object to act upon" can be reasonably defined for an
#      operation, it should be accepted as the first positional argument.
_OPERATIONS = KeyOrderedDict({
    # General operations - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    'define':       lambda d: d,
    'pass':         lambda d: d,
    'print':        print_data,

    'import':       import_module_or_object,
    'call':         lambda c, *a, **k: c(*a, **k),
    'import_and_call':
        lambda m, n, *a, **k: import_module_or_object(m, n)(*a, **k),

    # Some commonly used types
    'list':         list,
    'dict':         dict,
    'tuple':        tuple,
    'set':          set,

    'int':          int,
    'float':        float,
    'str':          str,
    
    # Item manipulation
    'getitem':      lambda d, k:    d[k],
    'setitem':      lambda d, k, v: d.__setitem__(k, v),

    # Attribute-related
    'getattr':      getattr,
    'setattr':      setattr,
    'callattr':     lambda d, attr, *a, **k: getattr(d, attr)(*a, **k),
    

    # Numerical operations - - - - - - - - - - - - - - - - - - - - - - - - - - 
    # Unary ...................................................................
    'increment':    lambda d: d + 1,
    'decrement':    lambda d: d - 1,
    'count_unique': count_unique,
    
    # numpy
    '.T':           lambda d: d.T,
    '.any':         lambda d: d.any(),
    '.all':         lambda d: d.all(),
    '.dtype':       lambda d: d.dtype,
    '.shape':       lambda d: d.shape,
    '.ndim':        lambda d: d.ndim,
    '.size':        lambda d: d.size,
    '.itemsize':    lambda d: d.itemsize,
    '.nbytes':      lambda d: d.nbytes,
    '.base':        lambda d: d.base,
    '.imag':        lambda d: d.imag,
    '.real':        lambda d: d.real,

    # xarray
    '.head':        lambda d: d.head(),
    '.tail':        lambda d: d.tail(),

    # logarithms and squares
    'log':          lambda d: np.log(d),
    'log10':        lambda d: np.log10(d),
    'log2':         lambda d: np.log2(d),
    'log1p':        lambda d: np.log1p(d),
    'squared':      lambda d: np.square(d),
    'sqrt':         lambda d: np.sqrt(d),
    'cubed':        lambda d: np.power(d, 3),
    'sqrt3':        lambda d: np.power(d, 1./.3),

    # Normalization and cumulation
    'normalize_to_sum':         lambda d: d / np.sum(d),
    'normalize_to_max':         lambda d: d / np.max(d),
    'cumulate':                 lambda d: np.cumsum(d),
    'cumulate_complementary':   lambda d: np.cumsum(d[::-1])[::-1],


    # Binary ..................................................................
    # Elementwise operations
    'add':          lambda d, v: operator.add(d, v),
    'concat':       lambda d, v: operator.concat(d, v),
    'div':          lambda d, v: operator.truediv(d, v),
    'truediv':      lambda d, v: operator.truediv(d, v),
    'floordiv':     lambda d, v: operator.floordiv(d, v),
    'lshift':       lambda d, v: operator.lshift(d, v),
    'mod':          lambda d, v: operator.mod(d, v),
    'mul':          lambda d, v: operator.mul(d, v),
    'matmul':       lambda d, v: operator.matmul(d, v),
    'rshift':       lambda d, v: operator.rshift(d, v),
    'sub':          lambda d, v: operator.sub(d, v),

    # numpy
    'power':        lambda d, e: np.power(d, e),

    # xarray
    '.coords':      lambda d, key: d.coords[key],


    # N-ary ...................................................................
    'create_mask':          create_mask,
    'where':                where,
    'populate_ndarray':     populate_ndarray,

    # dantro-specific wrappers around other library's functionality
    'dantro.multi_concat':  multi_concat,
    'dantro.merge':         merge,
    'dantro.expand_dims':   expand_dims,

    # numpy
    '.sum':         lambda d, **k: d.sum(**k),
    '.mean':        lambda d, **k: d.mean(**k),
    '.std':         lambda d, **k: d.std(**k),
    '.min':         lambda d, **k: d.min(**k),
    '.max':         lambda d, **k: d.max(**k),
    '.var':         lambda d, **k: d.var(**k),
    '.prod':        lambda d, **k: d.prod(**k),
    '.take':        lambda d, **k: d.take(**k),
    '.squeeze':     lambda d, **k: d.squeeze(**k),
    '.reshape':     lambda d, **k: d.reshape(**k),
    '.diagonal':    lambda d, **k: d.diagonal(**k),
    '.trace':       lambda d, **k: d.trace(**k),
    '.transpose':   lambda d, *a: d.transpose(*a),
    '.swapaxes':    lambda d, a1, a2: d.swapaxes(a1, a2),

    'invert':       lambda d, **k: np.invert(d, **k),
    'transpose':    lambda d, **k: np.transpose(d, **k),
    'diff':         lambda d, **k: np.diff(d, **k),
    'reshape':      lambda d, s, **k: np.reshape(d, s, **k),

    'np.array':     np.array,
    'np.empty':     np.empty,
    'np.zeros':     np.zeros,
    'np.ones':      np.ones,
    'np.arange':    np.arange,
    'np.linspace':  np.linspace,
    'np.logspace':  np.logspace,
    
    # xarray
    '.sel':         lambda d, **k: d.sel(**k),
    '.isel':        lambda d, **k: d.isel(**k),
    '.median':      lambda d, **k: d.median(**k),
    '.quantile':    lambda d, **k: d.quantile(**k),
    '.argmin':      lambda d, **k: d.argmin(**k),
    '.argmax':      lambda d, **k: d.argmax(**k),
    '.count':       lambda d, **k: d.count(**k),
    '.diff':        lambda d, **k: d.diff(**k),

    '.expand_dims':     lambda d, **k: d.expand_dims(**k),
    '.assign_coords':   lambda d, **k: d.assign_coords(**k),

    'xr.Dataset':   xr.Dataset,
    'xr.DataArray': xr.DataArray,
    'xr.merge':     xr.merge,
    'xr.concat':    xr.concat,
}) # End of default operation definitions

# Add the boolean operators
_OPERATIONS.update(BOOLEAN_OPERATORS)

# -----------------------------------------------------------------------------
# Registering and applying operations

def register_operation(*, name: str, func: Callable,
                       skip_existing: bool=False,
                       overwrite_existing: bool=False) -> None:
    """Adds an entry to the shared OPERATIONS registry.
    
    Args:
        name (str): The name of the operation
        func (Callable): The callable
        skip_existing (bool, optional): Description
        overwrite_existing (bool, optional): Description
    
    Raises:
        TypeError: On invalid name or non-callable for the func argument
        ValueError: On already existing operation name and no skipping or
            overwriting enabled.
    """
    if name in _OPERATIONS and not overwrite_existing:
        if skip_existing:
            return
        raise ValueError("Operation name '{}' already exists! Refusing to "
                         "register a new one. Set the overwrite_existing flag "
                         "to force overwriting.".format(name))

    elif not callable(func):
        raise TypeError("The given {} for operation '{}' is not callable! "
                        "".format(func, name))
    
    elif not isinstance(name, str):
        raise TypeError("Operation name need be a string, was {} with value "
                        "{}!".format(type(name), name))

    _OPERATIONS[name] = func

def apply_operation(op_name: str, *op_args, _log_level: int=5,
                    **op_kwargs) -> Any:
    """Apply an operation with the given arguments and then return it.
    
    Args:
        op_name (str): The name of the operation to carry out; need to be part
            of the OPERATIONS database.
        *op_args: The positional arguments to the operation
        _log_level (int, optional): Log level of the log messages created by
            this function.
        **op_kwargs: The keyword arguments to the operation
    
    Returns:
        Any: The result of the operation
    
    Raises:
        KeyError: On invalid operation name. This also suggests possible other
            names that might match.
        Exception: On failure to apply the operation, preserving the original
            exception.
    """
    try:
        op = _OPERATIONS[op_name]

    except KeyError as err:
        # Find some close matches to make operation discovery easier
        possible_matches = available_operations(match=op_name)

        raise ValueError("No operation '{}' registered! Did you mean: {} ?\n"
                         "Available operations:\n  - {}\n"
                         "If you need to register a new operation, use "
                         "dantro.utils.register_operation to do so."
                         "".format(op_name, ", ".join(possible_matches),
                                   "\n  - ".join(available_operations()))
                         ) from err

    # Compute and return the results
    log.log(_log_level, "Performing operation '%s' ...", op_name)
    try:
        return op(*op_args, **op_kwargs)

    except Exception as exc:
        raise RuntimeError("Failed applying operation '{}'! Got a {}: {}\n"
                           "  args:   {}\n"
                           "  kwargs: {}\n"
                           "".format(op_name, exc.__class__.__name__, str(exc),
                                     op_args, op_kwargs)
                           ) from exc


def available_operations(*, match: str=None, n: int=5) -> Sequence[str]:
    """Returns all available operation names or a fuzzy-matched subset of them.
    
    Args:
        match (str, optional): If given, fuzzy-matches the names and only
            returns close matches to this name.
        n (int, optional): Number of close matches to return. Passed on to
            difflib.get_close_matches
    
    Returns:
        Sequence[str]: All available operation names or the matched subset.
            The sequence is sorted alphabetically.
    """
    if match is None:
        return _OPERATIONS.keys()

    # Use fuzzy matching to return close matches
    return get_close_matches(match, _OPERATIONS.keys(), n=n)



"""This sub-package provides mixin classes for easily adding functionality
to a derived contaier or group class
"""

from .base import AttrsMixin, SizeOfMixin, CheckDataMixin, LockDataMixin
from .base import CollectionMixin, ItemAccessMixin, MappingAccessMixin

from .general import ForwardAttrsMixin, ForwardAttrsToDataMixin
from .indexing import IntegerItemAccessMixin, PaddedIntegerItemAccessMixin
from .numeric import UnaryOperationsMixin, NumbersMixin, ComparisonMixin
from .proxy_support import ProxySupportMixin, Hdf5ProxySupportMixin



"""This submodule contains some utility classes and functions"""

from .ordereddict import KeyOrderedDict, IntOrderedDict
from .link import Link
from .coords import (extract_dim_names, extract_coords,
                     extract_coords_from_attrs, extract_coords_from_name,
                     extract_coords_from_data)
from .data_ops import register_operation, apply_operation, available_operations



"""This module implements the PlotHelper class"""

import os
import copy
import logging
import inspect
from itertools import product
from typing import Union, Callable, Tuple, List, Dict, Generator

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from dantro.tools import recursive_update

# Public constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Tools

class temporarily_changed_axis:
    """Context manager to temporarily change an axis in the PlotHelper"""

    def __init__(self, hlpr, *, tmp_ax_coords: Tuple[int]=None):
        """Initialize the context manager.

        Args:
            hlpr: The plot helper of which to select a temporary axis
            tmp_ax_coords (Tuple[int]): The coordinates of the temporary axis.
                If not given, will not change the axis.
        """
        self._hlpr = hlpr
        self._tmp_ax_coords = tmp_ax_coords
        self._old_ax_coords = None

    def __enter__(self):
        """Enter the context, selecting a temporary axis"""
        # Store the current axis' coordinates
        self._old_ax_coords = self._hlpr.ax_coords
        
        # If it needs to be changed, do it.
        if (    self._tmp_ax_coords is not None
            and self._tmp_ax_coords != self._old_ax_coords):
            log.debug("Temporarily changing from axis %s to %s ...",
                      self._old_ax_coords, self._tmp_ax_coords)
            self._hlpr.select_axis(*self._tmp_ax_coords)
        
        else:
            log.debug("No need to change current axis.")

    def __exit__(self, *args):
        """Change back to the initial axis. Errors are not handled."""
        if self._old_ax_coords != self._hlpr.ax_coords:
            log.debug("Changing back to axis %s ...", self._old_ax_coords)
            self._hlpr.select_axis(*self._old_ax_coords)

def coords_match(coords: Tuple[int], *,
                 match: Union[tuple, str], full_shape: Tuple[int]) -> bool:
    """Whether a coordinate is matched by a coordinate match tuple.

    Allowed values in the coordinate match tuple are:
        * integers: regarded as coordinates. If negative or exceeding the full
            shape, these are wrapped around.
        * Ellipsis: matches all coordinates
        * None: alias for Ellipsis
    
    Args:
        coords (Tuple[int]): The coordinate to match
        match (Union[tuple, str]): The match tuple, where None is
            interpreted as an Ellipsis and negative values are wrapped around
            by `full_shape`. Can also be 'all', which is equivalent to a
            (None, None) tuple. Can also be a list, which is then converted to
            a tuple.
        full_shape (Tuple[int]): The full shape of the axes; needed to wrap
            around negative values in `match`.
    
    Returns:
        bool: Whether `coords` matches `match`
    
    Raises:
        TypeError: `match` not being a tuple or a list
        ValueError: Any of the arguments not being 2-tuples.
    """
    # Convert the 'all argument'
    match = match if match != 'all' else (Ellipsis, Ellipsis)

    # Make sure it is a tuple, allowing conversion from lists
    if isinstance(match, list):
        match = tuple(match)
    
    elif not isinstance(match, tuple):
        raise TypeError("Argument `match` needs to be a 2-tuple, list, or a "
                        "string, but was {} with value '{}'!"
                        "".format(type(match), match))

    # Convert any Nones to Ellipsis
    match = tuple([m if m is not None else Ellipsis for m in match])
    
    # Check length and values, not allowing values exceeding the shape
    if any([len(t) != 2 for t in (coords, match, full_shape)]):
        raise ValueError("Need 2-tuples for arguments, got {}, {}, and {}!"
                         "".format(coords, match, full_shape))

    elif not all([m is Ellipsis or m < s for m, s in zip(match, full_shape)]):
        raise ValueError("Got match values {} exceeding the shape {}! Take "
                         "care that all values are strictly smaller than the "
                         "maximum value. Negative values are allowed and will "
                         "be evaluated via a modulo operation."
                         "".format(match, full_shape))

    for c, m, s in zip(coords, match, full_shape):
        if m is Ellipsis:
            # Always matches
            continue
        if (m % s) != c:
            # No match
            return False
        
    # Went through -> have a match
    return True


# -----------------------------------------------------------------------------

class PlotHelper:
    """The PlotHelper takes care of the figure setup and saving and allows
    accessing matplotlib utilities through the plot configuration.
    """

    # Configuration keys with special meaning
    _SPECIAL_CFG_KEYS = ('setup_figure', 'save_figure')

    def __init__(self, *, out_path: str, helper_defaults: dict=None,
                 update_helper_cfg: dict=None, raise_on_error: bool=True):
        """Initialize a Plot Helper with a certain configuration.
        
        This configuration is the so-called "base" configuration and is not
        axis-specific. There is the possibility to specify axis-specific
        configuration entries.
        
        All entries in the helper configuration are deemed 'enabled' unless
        they explicitly specify `enabled: false` in their configuration.
        
        Args:
            out_path (str): path to store the created figure
            helper_defaults (dict, optional): The basic configuration of the
                helpers.
            update_helper_cfg (dict, optional): A configuration used to update
                the existing helper defaults
            raise_on_error (bool, optional): Whether to raise on an exception
                created on helper invocation or just log the error
        """
        # Determine available helper methods, store it as tuple
        self._AVAILABLE_HELPERS = [attr_name[6:] for attr_name in dir(self)
                                   if attr_name.startswith('_hlpr_')
                                   and callable(getattr(self, attr_name))]
        self._AVAILABLE_HELPERS = tuple(self._AVAILABLE_HELPERS)

        # Store (a copy of) the base configuration
        self._base_cfg = (copy.deepcopy(helper_defaults) if helper_defaults
                          else {})

        # Update the defaults
        if update_helper_cfg:
            self._base_cfg = recursive_update(self._base_cfg,
                                              copy.deepcopy(update_helper_cfg))

        # Extract the axis-specific update list
        self._axis_specific_updates = self._base_cfg.pop('axis_specific', {})

        # Check that all remaining entries are valid keys
        self._raise_on_invalid_helper_name(*self._base_cfg.keys(),
                                           special_cfg_keys_allowed=True)

        # Initialize the actual axis-specific configuration empty; it can only
        # be compiled once the figure is created.
        self._cfg = None

        # Store the other attributes
        self._out_path = out_path
        self._raise_on_error = raise_on_error

        # Initialize attributes that are set at a later point
        self._fig = None
        self._axes = None
        self._current_ax_coords = None
        self._additional_axes = None
        self._animation_update = None

    # .........................................................................
    # Properties

    @property
    def _axis_cfg(self) -> dict:
        """Return the configuration for the current axis; not a deep copy!"""
        try:
            return self._cfg[self.ax_coords]
        
        except RuntimeError as err:
            raise RuntimeError("The axis-specific configuration is only "
                               "available while a figure is associated with "
                               "the PlotHelper!") from err

    @property
    def axis_cfg(self) -> dict:
        """Returns a deepcopy of the current axis' configuration."""
        return copy.deepcopy(self._axis_cfg)
    
    @property
    def base_cfg(self) -> dict:
        """Returns a deepcopy of the base configuration, i.e. the configuration
        that is not axis-specific.
        """
        return copy.deepcopy(self._base_cfg)

    @property
    def fig(self):
        """Returns the current figure"""
        if self._fig is None:
            raise ValueError("No figure initialized or already closed! Use "
                             "the `setup_figure` method to create a figure "
                             "instance.")
        return self._fig

    @property
    def ax(self):
        """Returns the current axis of the associated figure"""
        return self.fig.gca()
    
    @property
    def ax_coords(self) -> Tuple[int]:
        """Returns the current axis coordinates within a subfigure in shape
        ``(col, row)``.

        For example, the ``(0, 0)`` coordinate refers to the top left subplot
        of the figure. ``(1, 2)`` is the axis object in the second column,
        third row.
        """
        if self._current_ax_coords is None:
            raise RuntimeError("The current axis coordinate is only defined "
                               "while a figure is associated with the "
                               "PlotHelper!")

        return self._current_ax_coords
        # NOTE There _would_ be the possiblity to use the matplotlib axis
        #      properties self.ax.numCol and .numRow, which store the column
        #      and row the axis was created in via plt.subplots. However, that
        #      that information gets lost as soon as an additional axis is
        #      added to the figure, e.g. when adding a color bar. Thus, the
        #      current axis coordinates need to be stored in an attribute.

    @property
    def axes(self) -> np.ndarray:
        """Returns the axes array, which is of shape (#cols, #rows).

        The (0, 0) axis refers to the top left subplot of the figure.
        """
        return self._axes
    
    @property
    def available_helpers(self) -> Tuple[str]:
        """List of available helper names"""
        return self._AVAILABLE_HELPERS

    @property
    def enabled_helpers(self) -> list:
        """Returns a list of enabled helpers for the current axis"""
        return [hn for hn in self._axis_cfg
                if self._axis_cfg[hn].get('enabled', True)
                and hn not in self._SPECIAL_CFG_KEYS]

    @property
    def out_path(self) -> str:
        """Returns the output path of the plot"""
        return self._out_path

    @property
    def animation_update(self) -> Callable:
        """Returns the animation update generator callable"""
        if self._animation_update is None:
            raise ValueError("No animation update generator was registered "
                             "with the PlotHelper! Cannot perform animation "
                             "update.")
        return self._animation_update
    

    # .........................................................................
    # Figure setup and axis control

    def attach_figure_and_axes(self, *, fig, axes):
        """Attaches the given figure and axes to the PlotHelper. This method
        replaces an existing figure and existing axes with the ones given.

        As the PlotHelper relies on axes being accessible via coordinate pairs,
        multiple axes must be passed as two-dimensional array-like. Since the
        axes are internally stored as numpy array, the axes-grid must be
        complete.
        
        Note that by closing the old figure the existing axis-specific config
        and all existing axes are destroyed. In other words: All information
        previously provided via the provide_defaults and the mark_* methods is
        lost. Therefore, if needed, it is recommended to call this method at
        the beginning of the plotting function.

        .. note::

            This function assumes multiple axes to be passed in (y,x) format
            (as e.g. returned by matplotlib.pyplot.subplots with squeeze set to
            False) and internally transposes the axes-grid such that afterwards
            it is accessible via (x,y) coordinates.
        
        Args:
            fig: The new figure which replaces the existing.
            axes: single axis or 2d array-like containing the axes
        
        Raises:
            ValueError: On multiple axes not being passed in 2d format.
        
        """
        if self._fig is not None:
            log.debug("Closing existing figure and re-associating with a new "
                      "figure ...")
            self.close_figure()

        # Assign the new figure
        self._fig = fig

        # Prepare the new axis object
        try:
            # Assuming it's a scalar, np.reshape leads to np.array being called
            # on the object, thus allowing any scalar type. Only in cases where
            # an np.ndarray with size > 1 is given will this reshape operation
            # fail.
            axes = np.reshape(axes, (1, 1))

        except ValueError:
            # Else, assume array-like, containing the axes in (y,x) format.
            # Transpose the axes such that they are accessible in the (x,y)
            # format, which is used internally throughout the PlotHelper.
            axes = np.array(axes).T

        # Ensure correct shape
        if axes.ndim != 2:
            raise ValueError("When attaching a figure with multiple axes, the "
                             "axes must be passed as a 2d array-like object! "
                             "Got object of shape {}.".format(axes.shape))

        # Everything ok, attach axes
        self._axes = axes

        log.debug("Figure and axes attached.")

        # Select the (0, 0) axis, for consistency
        self.select_axis(0, 0)

        # Can now evaluate the axis-specific configuration
        self._cfg = self._compile_axis_specific_cfg()


    def setup_figure(self, **update_fig_kwargs):
        """Sets up a matplotlib figure instance and axes with the given
        configuration (by calling matplotlib.pyplot.subplots) and attaches
        both to the PlotHelper.

        If the ``scale_figsize_with_subplots_shape`` option is enabled here,
        this method will also take care of scaling the figure accordingly.
        
        Args:
            **update_fig_kwargs: Parameters that are used to update the
                figure setup parameters stored in `setup_figure`.
        """
        # Prepare arguments
        fig_kwargs = self.base_cfg.get('setup_figure', {})

        if update_fig_kwargs:
            fig_kwargs = recursive_update(fig_kwargs, update_fig_kwargs)

        # Need to handle scaling argument separately
        scale_figsize = fig_kwargs.pop('scale_figsize_with_subplots_shape',
                                       False)

        # Now, create the figure and axes and attach them
        fig, axes = plt.subplots(squeeze=False, **fig_kwargs)
        log.debug("Figure created.")

        self.attach_figure_and_axes(fig=fig, axes=axes)

        # Scale figure, if needed
        if scale_figsize and self.axes.size > 1:
            log.debug("Scaling current figure size with subplots shape %s ...",
                      self.axes.shape)

            old_figsize = self.fig.get_size_inches()
            self.fig.set_size_inches(old_figsize[0] * self.axes.shape[0],
                                     old_figsize[1] * self.axes.shape[1])

            log.debug("Scaled figure size from %s to %s.",
                      old_figsize, self.fig.get_size_inches())

    def save_figure(self, *, close: bool=True):
        """Saves and (optionally, but default) closes the current figure
        
        Args:
            close (bool, optional): Whether to close the figure after saving.
        """
        self.fig.savefig(self.out_path, **self.base_cfg.get('save_figure', {}))
        log.debug("Figure saved.")

        if close:
            self.close_figure()

    def close_figure(self):
        """Closes the figure and disassociates it from the helper.

        This also removes the axes objects and deletes the axis-specific
        configuration. All information provided via provide_defaults and the
        mark_* methods is lost.
        """
        plt.close(self.fig)
        log.debug("Figure closed.")

        self._fig = None
        self._axes = None
        self._current_ax_coords = None
        self._cfg = None
        log.debug("Associated data removed.")
    
    def select_axis(self, col: int, row: int):
        """Selects the axes at the given coordinate as the current axis.

        This does not perform a check on whether the axis is valid or already
        set.
        
        Args:
            col (int): The column to select, i.e. the x-coordinate. Can be
                negative, in which case it indexes backwards from the last
                column.
            row (int): The row to select, i.e. the y-coordinate. Can be
                negative, in which case it indexes backwards from the last row.
        
        Raises:
            ValueError: On failing to set the current axis
        
        """
        log.debug("Selecting axis (%s, %s) ...", col, row)

        try:
            self.fig.sca(self.axes[col, row])

        except IndexError as exc:
            raise ValueError("Could not select axis ({}, {}) from figure with "
                             "subplots of shape {}!"
                             "".format(col, row, self.axes.shape)) from exc
        
        else:
            self._current_ax_coords = (col % self.axes.shape[0],
                                       row % self.axes.shape[1])

    def coords_iter(self, *, match: Union[tuple, str]=None,
                    ) -> Generator[Tuple[int], None, None]:
        """Returns a generator to iterate over all coordinates that match
        `match`.
        
        Args:
            match (Union[tuple, str]): The coordinates to match; those that do
                not match this pattern (evaluated by `coords_match` function)
                will not be yielded. If not given, will iterate only over the
                currently selected axis.
        
        Yields:
            Generator[Tuple[int], None, None]: The axis coordinates generator
        """
        # If no match argument is given, match only current coordinates
        match = match if match is not None else self.ax_coords

        # Go over the cartesian product ranges specified by the subplots shape
        for coords in product(range(self.axes.shape[0]),
                              range(self.axes.shape[1])):
            if coords_match(coords, match=match, full_shape=self.axes.shape):
                yield coords

    # .........................................................................
    # Helper invocation and configuration

    def _invoke_helper(self, helper_name: str, *,
                       axes: Union[tuple, str]=None,
                       mark_disabled_after_use: bool=True,
                       **update_kwargs) -> None:
        """Invokes a single helper on the specified axes, if it is enabled, and
        marks it disabled afterwards. The given update parameters are used to
        update the existing configuration.
        
        Unlike the public invoke_helper method, this method checks for whether
        the helper is enabled, while the public method automatically assumes
        it is meant to be enabled.
        
        Args:
            helper_name (str): helper which is invoked
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): If True, the helper is
                marked as disabled after invoking it
            **update_kwargs: Update parameters for this specific plot helper.
                Note that these do not persist, but are only used for this
                invocation.
        
        Raises:
            ValueError: No matching helper function defined
        
        Returns:
            None
        """
        # Get the helper function
        try:
            helper = getattr(self, "_hlpr_" + helper_name)

        except AttributeError as err:
            raise ValueError("No helper with name '{}' available! "
                             "Available helpers: {}"
                             "".format(helper_name,
                                       ", ".join(self.available_helpers))
                             ) from err

        # Go over all matching axis coordinates
        for ax_coords in self.coords_iter(match=axes):
            # Temporarily change to this axis
            with temporarily_changed_axis(self, tmp_ax_coords=ax_coords):
                # Prepare the helper params for this axis, working on a copy
                helper_params = self.axis_cfg.get(helper_name, {})
                
                if update_kwargs:
                    recursive_update(helper_params,
                                     copy.deepcopy(update_kwargs))

                # If it is disabled, go to the next iteration
                if not helper_params.pop('enabled', True):
                    log.debug("Helper '%s' is not enabled for axis %s.",
                              helper_name, self.ax_coords)
                    continue
                
                # Invoke helper
                log.debug("Invoking helper function '%s' on axis %s ...",
                          helper_name, self.ax_coords)

                try:
                    helper(**helper_params)

                except Exception as exc:
                    # Build an informative error message
                    hp_params = "\n".join(["   {}: {}".format(k, repr(v))
                                           for k, v in helper_params.items()])
                    hp_doc = inspect.getdoc(helper)

                    msg = ("A {} was raised during invocation of the '{}' "
                           "helper: {}.\n\nIt was invoked with the following "
                           "arguments:\n{}\n\nMake sure these arguments were "
                           "valid. You may want to consult the helper's "
                           "docstring:\n\n{}"
                           "".format(exc.__class__.__name__, helper_name, exc,
                                     hp_params, hp_doc))

                    # Either log or raise
                    if self._raise_on_error:
                        raise ValueError(msg) from exc
                    log.error(msg)

                if mark_disabled_after_use:
                    self.mark_disabled(helper_name)

            # Now back at previous axis, whatever happened above.

    def invoke_helper(self, helper_name: str, *,
                      axes: Union[tuple, str]=None,
                      mark_disabled_after_use: bool=True,
                      **update_kwargs):
        """Invokes a single helper on the specified axes.
        
        Args:
            helper_name (str): The name of the helper to invoke
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): If True, the helper is
                marked as disabled after the invocation.
            **update_kwargs: Update parameters for this specific plot helper.
                Note that these do not persist, but are only used for this
                invocation.
        """
        self._invoke_helper(helper_name, axes=axes, enabled=True,
                            mark_disabled_after_use=mark_disabled_after_use,
                            **update_kwargs)

    def invoke_helpers(self, *helper_names,
                       axes: Union[tuple, str]=None,
                       mark_disabled_after_use: bool=True,
                       **update_helpers):
        """Invoke all specified helpers on the specified axes.
        
        Args:
            *helper_names: The helper names to invoke
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): Whether to mark helpers
                disabled after they were used.
            **update_helpers: Update parameters for all plot helpers.
                These have to be grouped under the name of the helper in order
                to be correctly associated. Note that these do not persist,
                but are only used for this invocation.
        
        Deleted Parameters:
            helper_names (list): Helpers to be invoked
        """
        for helper_name in helper_names:
            self.invoke_helper(helper_name, axes=axes,
                               mark_disabled_after_use=mark_disabled_after_use,
                               **update_helpers.get(helper_name, {}))

    def invoke_enabled(self, *,
                       axes: Union[tuple, str]=None,
                       **update_helpers):
        """Invokes all enabled helpers with their current configuration on the
        matching axes.
        
        Args:
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            **update_helpers: Update parameters for all plot helpers.
                These have to be grouped under the name of the helper in order
                to be correctly associated. Note that these do not persist,
                but are only used for this invocation.
        """
        # Go over all axes matching the given argument
        for ax_coords in self.coords_iter(match=axes):
            # Temporarily change to this axis
            with temporarily_changed_axis(self, tmp_ax_coords=ax_coords):
                # Iterate over all enabled helpers on this axis
                for helper_name in self.enabled_helpers:
                    # See if there are update parameters for this helper
                    params = update_helpers.get(helper_name, {})

                    # Invoke the single helper on the current axis
                    self._invoke_helper(helper_name, **params)

    def provide_defaults(self, helper_name: str, *,
                         axes: Union[tuple, str]=None,
                         mark_enabled: bool=True, **update_kwargs):
        """Update or add a single entry to a helper's configuration.
        
        Args:
            helper_name (str): The name of the helper whose config is to change
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_enabled (bool, optional): Whether to mark the helper enabled
                by providing defaults
            **update_kwargs: dict containing the helper parameters with
                which the config is updated recursively
        """
        self._raise_on_invalid_helper_name(helper_name,
                                           special_cfg_keys_allowed=True)

        # Special configuration keys are updating the base configuration
        if helper_name in self._SPECIAL_CFG_KEYS:
            if helper_name not in self._base_cfg:
                self._base_cfg[helper_name] = dict()

            # Do an in-place recursive update; can return afterwards
            recursive_update(self._base_cfg[helper_name],
                             copy.deepcopy(update_kwargs))
            log.debug("Updated defaults for special configuration entry '%s'.",
                      helper_name)
            return

        # Go over all selected axes
        for ax_coords in self.coords_iter(match=axes):
            # Update or create the configuration entry, manually passing the
            # coordinates. Using the _axis_cfg property would require to enter
            # the temporary axis context, which is overkill at this point.
            if helper_name not in self._cfg[ax_coords]:
                # Create a new empty entry
                self._cfg[ax_coords][helper_name] = dict()

            # Can do an in-place recursive update now
            recursive_update(self._cfg[ax_coords][helper_name],
                             copy.deepcopy(update_kwargs))
            log.debug("Updated axis-specific defaults for helper '%s' and "
                      "axis %s.", helper_name, ax_coords)

            if mark_enabled and helper_name not in self._SPECIAL_CFG_KEYS:
                self.mark_enabled(helper_name, axes=ax_coords)

    def mark_enabled(self, *helper_names, axes: Union[tuple, str]=None):
        """Marks the specified helpers as enabled for the specified axes.
        
        Args:
            *helper_names: Helpers to be enabled
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
        """
        # Check the helper names
        self._raise_on_invalid_helper_name(*helper_names)

        # Go over all selected axes
        for ax_coords in self.coords_iter(match=axes):
            # Go over the specified helper names
            for helper_name in helper_names:
                # Create the empty dict, if necessary
                if helper_name not in self._cfg[ax_coords]:
                    self._cfg[ax_coords][helper_name] = dict()

                self._cfg[ax_coords][helper_name]['enabled'] = True
                log.debug("Marked helper '%s' enabled for axis %s.",
                          helper_name, ax_coords)

    def mark_disabled(self, *helper_names, axes: Union[tuple, str]=None):
        """Marks the specified helpers as disabled for the specified axes.
        
        Args:
            *helper_names: Helpers to be disabled
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
        """
        # Check the helper names
        self._raise_on_invalid_helper_name(*helper_names)

        # Go over all selected axes
        for ax_coords in self.coords_iter(match=axes):
            # Go over the specified helper names
            for helper_name in helper_names:
                # Create the empty dict, if necessary
                if helper_name not in self._cfg[ax_coords]:
                    self._cfg[ax_coords][helper_name] = dict()

                self._cfg[ax_coords][helper_name]['enabled'] = False
                log.debug("Marked helper '%s' disabled for axis %s.",
                          helper_name, ax_coords)

    def register_animation_update(self, update_func: Callable):
        """Registers a generator used for animations.
        
        Args:
            update_func (Callable): Generator object over which is iterated
                over to create an animation. This needs
        """
        self._animation_update = update_func
        log.debug("Registered animation update function.")

    # .........................................................................
    # Private support methods

    def _compile_axis_specific_cfg(self) -> Dict[tuple, dict]:
        """With a figure set up, compiles the axis-specific helper config."""
        # The target dict of axis coordinates to fully resolved dicts
        cfg = dict()

        # Compile a base configuration dict that does not contain special keys
        base_cfg = {k: v for k, v in self._base_cfg.items()
                    if k not in self._SPECIAL_CFG_KEYS}
        # NOTE setup_figure and save_figure are handled separately

        # Go over all possible coordinates
        for ax_coords in self.coords_iter(match='all'):
            # Store a copy of the base configuration for these coordinates
            cfg[ax_coords] = copy.deepcopy(self._base_cfg)

            # Go over the list of updates and apply them
            for update_params in self._axis_specific_updates.values():
                # Work on a copy
                update_params = copy.deepcopy(update_params)

                # Extract the axis to update
                axis = update_params.pop('axis')

                # Check if there is a match
                if not coords_match(ax_coords, match=axis,
                                    full_shape=self.axes.shape):
                    # Nothing to update for this coordinate
                    continue

                # In-place update the parameter dict. Copies have been created
                # above already, so no need to do this here again.
                recursive_update(cfg[ax_coords], update_params)

        # Now created configurations for all axes and applied all updates
        return cfg

    def _raise_on_invalid_helper_name(self, *helper_names: str,
                                      special_cfg_keys_allowed: bool=False):
        """Makes sure the given helper names are valid.
        
        Args:
            *helper_names (str): Names of the helpers to check
            special_cfg_keys_allowed (bool, optional): Whether to regard the
                special configuration keys as valid or not.
        
        Raises:
            ValueError: On invalid helper name
        """
        valid_names = self.available_helpers

        if special_cfg_keys_allowed:
            valid_names += self._SPECIAL_CFG_KEYS

        for helper_name in helper_names:
            if helper_name in valid_names:
                # Is valid
                continue

            # Not valid!
            raise ValueError("No helper with name '{}' available! "
                             "Available helpers: {}. "
                             "Special configuration keys ({}) {} allowed for "
                             "this operation."
                             "".format(helper_name,
                                       ", ".join(self.available_helpers),
                                       ", ".join(self._SPECIAL_CFG_KEYS),
                                       ("are also" if special_cfg_keys_allowed
                                        else "are NOT")))

    # .........................................................................
    # Helper Methods -- acting on the figure

    def _hlpr_set_suptitle(self, *, title: str=None, **title_kwargs):
        """Set the figure title, i.e. matplotlib.Figure.suptitle
        
        Args:
            title (str, optional): The title to be set
            **title_kwargs: Passed on to plt.set_title
        """
        if title:
            self.fig.suptitle(title, **title_kwargs)


    # .........................................................................
    # Helper Methods -- acting on a single axis

    def _hlpr_set_title(self, *, title: str=None, **title_kwargs):
        """Set the title of the current axis
        
        Args:
            title (str, optional): The title to be set
            **title_kwargs: Passed on to plt.set_title
        """
        if title:
            self.ax.set_title(title, **title_kwargs)

    def _hlpr_set_labels(self, *,
                         x: Union[str, dict]=None,
                         y: Union[str, dict]=None,
                         only_label_outer: bool=False):
        """Set the x and y label of the current axis
        
        Args:
            x (Union[str, dict], optional): Either the label as a string or
                a dict with key `label`, where all further keys are passed on
                to plt.set_xlabel
            y (Union[str, dict], optional): Either the label as a string or
                a dict with key `label`, where all further keys are passed on
                to plt.set_ylabel
            only_label_outer (bool, optional): Whether to label only outer axes
        """

        def set_label(func: Callable, *, label: str=None, **label_kwargs):
            # NOTE Can be extended here in the future to do more clever things
            return func(label, label_kwargs)

        if x:
            x = x if not isinstance(x, str) else dict(label=x)
            set_label(self.ax.set_xlabel, **x)

        if y:
            y = y if not isinstance(y, str) else dict(label=y)
            set_label(self.ax.set_ylabel, **y)

        if only_label_outer:
            self.ax.label_outer()

    def _hlpr_set_limits(self, *,
                         x: Union[tuple, dict]=None,
                         y: Union[tuple, dict]=None):
        """Set the x and y limit for the current axis
        
        x and y can have the following shapes:
            None           Limits are not set
            tuple, list    Specify lower and upper values
            dict           expecting keys `lower` and/or `upper`

        Each entries of the tuple or dict values can be:
            None           Set automatically / do not set
            numeric        Set to this value explicitly
            min            Set to the data minimum value on that axis
            max            Set to the data maximum value on that axis
        
        Args:
            x (Union[tuple, dict], optional): Set the x-axis limits. For valid
                argument values, see above.
            y (Union[tuple, dict], optional): Set the y-axis limits. For valid
                argument values, see above.
        """
        def parse_args(args: Union[tuple, dict], *, ax):
            """Parses the limit arguments."""

            def parse_arg(arg: Union[float, str]) -> Union[float, None]:
                """Parses a single limit argument to either be float or None"""
                if not isinstance(arg, str):
                    # Nothing to parse
                    return arg

                if arg == 'min':
                    arg = ax.get_data_interval()[0]
                elif arg == 'max':
                    arg = ax.get_data_interval()[1]
                else:
                    raise ValueError("Got an invalid str-type argument '{}' "
                                     "to set_limits helper. Allowed: min, max."
                                     "".format(arg))

                # Check that it is finite
                if not np.isfinite(arg):
                    raise ValueError("Could not get a finite value from the "
                                     "axis data to use for setting axis "
                                     "limits to 'min' or 'max', presumably "
                                     "because the axis is still empty.")

                return arg

            # Special case: dict
            if isinstance(args, dict):
                # Make sure there are only allowed keys
                if [k for k in args.keys() if k not in ('lower', 'upper')]:
                    raise ValueError("There are invalid keys present in a "
                                     "dict-type argument to set_limits! Only "
                                     "accepting keys 'lower' and 'upper', but "
                                     "got: {}".format(args))

                # Unpack into tuple
                args = (args.get('lower', None), args.get('upper', None))


            # Make sure it is a list or tuple of size 2
            if not isinstance(args, (tuple, list)):
                raise TypeError("Argument for set_limits helper needs to be "
                                "a dict, list, or a tuple, but was of type {} "
                                "with value '{}'!"
                                "".format(type(args), args))

            if len(args) != 2:
                raise ValueError("Argument for set_limits helper needs to be "
                                 "a list or tuple of length 2 or a dict with "
                                 "keys 'upper' and/or 'lower', but was {}!"
                                 "".format(args))

            # Parse and return
            return (parse_arg(args[0]), parse_arg(args[1]))

        # Now set the limits, using the helper functions defined above
        if x is not None:
            self.ax.set_xlim(*parse_args(x, ax=self.ax.xaxis))
        
        if y is not None:
            self.ax.set_ylim(*parse_args(y, ax=self.ax.yaxis))

    def _hlpr_set_legend(self, *, use_legend: bool=True, **legend_kwargs):
        """Set a legend for the current axis"""
        if use_legend:
            handles, labels = self.ax.get_legend_handles_labels()
            self.ax.legend(handles, labels, **legend_kwargs)

    def _hlpr_set_texts(self, *, texts: list):
        """Set a list of texts for the current axis
        
        Args:
            texts: The list of text dicts, that are passed to 
            matplotlib.pyplot.text
        """
        for text_args in texts:
            self.ax.text(**text_args)

    def _hlpr_set_hv_lines(self, *, hlines: list=None, vlines: list=None):
        """Set one or multiple horizontal or vertical lines.
        
        Args:
            hlines (list, optional): list of numeric positions of the lines or
                or list of dicts with key `pos` determining the position, key
                `limits` determining the relative limits of the line, and all
                additional arguments being passed on to the matplotlib
                function.
            vlines (list, optional): list of numeric positions of the lines or
                or list of dicts with key `pos` determining the position, key
                `limits` determining the relative limits of the line, and all
                additional arguments being passed on to the matplotlib
                function.
        """

        def set_line(func: Callable, *, pos: float, limits: tuple=(0., 1.),
                     **line_kwargs):
            """Helper function to invoke the matplotlib function that sets 
            a horizontal or vertical line."""
            try:
                pos = float(pos)
            
            except Exception as err:
                raise ValueError("Got non-numeric value '{}' for `pos` "
                                 "argument in set_hv_lines helper!"
                                 "".format(pos))

            func(pos, *limits, **line_kwargs)

        if hlines is not None:
            for line_spec in hlines:
                if isinstance(line_spec, dict):
                    set_line(self.ax.axhline, **line_spec)
                else:
                    set_line(self.ax.axhline, pos=line_spec)
        
        if vlines is not None:
            for line_spec in vlines:
                if isinstance(line_spec, dict):
                    set_line(self.ax.axvline, **line_spec)
                else:
                    set_line(self.ax.axvline, pos=line_spec)

    def _hlpr_set_scales(self, *,
                         x: Union[str, dict]=None,
                         y: Union[str, dict]=None):
        """Set a scale for the current axis"""

        def set_scale(func: Callable, *, scale: str=None, **scale_kwargs):
            func(scale, **scale_kwargs)

        if x:
            x = x if not isinstance(x, str) else dict(scale=x)
            set_scale(self.ax.set_xscale, **x)

        if y:
            y = y if not isinstance(y, str) else dict(scale=y)
            set_scale(self.ax.set_yscale, **y)



"""This module implements specialisations of the BaseDataContainer class."""

import logging
from typing import Union, Tuple, Sequence

import numpy as np
import xarray as xr
import copy

from ..base import BaseDataContainer, ItemAccessMixin, CheckDataMixin
from ..mixins import ForwardAttrsToDataMixin, NumbersMixin, ComparisonMixin
from ..abc import AbstractDataProxy
from ..utils import Link, extract_dim_names, extract_coords

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class XrDataContainer(ForwardAttrsToDataMixin, NumbersMixin, ComparisonMixin,
                      CheckDataMixin, ItemAccessMixin, BaseDataContainer):
    """The XrDataContainer stores numerical xarray.DataArray data associated
    with dimensions, coordinates, and attributes.
    """
    # Specify expected data types for this container class
    DATA_EXPECTED_TYPES = (xr.DataArray, np.ndarray,)
    DATA_ALLOW_PROXY = False
    DATA_UNEXPECTED_ACTION = 'raise'

    # Custom class variables for customizing XrDataContainer ..................
    # Define as class variable the name of the attribute that determines the
    # dimensions of the xarray.DataArray
    _XRC_DIMS_ATTR = 'dims'
    
    # Attributes prefixed with this string can be used to set names for
    # specific dimensions. The prefix should be followed by an integer-parsable
    # string, e.g. `dim_name__0` would be the dimension name for the 0th dim.
    _XRC_DIM_NAME_PREFIX = 'dim_name__'

    # Attributes prefixed with this string determine the coordinate values for
    # a specific dimension. The prefix should be followed by the _name_ of the
    # dimension, e.g. `coord__time`. The values are interpreted according to
    # the default coordinate mode or, if given, the coord_mode__* attribute
    _XRC_COORDS_ATTR_PREFIX = 'coords__'

    # The default mode by which coordinates are interpreted
    _XRC_COORDS_MODE_DEFAULT = 'values'

    # Prefix for the coordinate mode if a custom mode is to be used
    _XRC_COORDS_MODE_ATTR_PREFIX = 'coords_mode__'

    # Whether to inherit the other container attributes
    _XRC_INHERIT_CONTAINER_ATTRIBUTES = True

    # Whether to use strict attribute checking; throws errors if there are
    # container attributes available that match the prefix but don't match a
    # valid dimension name. Can be disabled for speed improvements
    _XRC_STRICT_ATTR_CHECKING = True

    # .........................................................................

    def __init__(self, *, name: str, data: Union[np.ndarray, xr.DataArray],
                 dims: Sequence[str]=None, coords: dict=None,
                 extract_metadata: bool=True, apply_metadata: bool=True,
                 **dc_kwargs):
        """Initialize a XrDataContainer and extract dimension and coordinate
        labels.
        
        Args:
            name (str): which name to give to the XrDataContainer
            data (Union[np.ndarray, xr.DataArray]): The data to store; anything
                that an xr.DataArray can take
            dims (Sequence[str], optional): The dimension names.
            coords (dict, optional): The coordinates. The keys of this dict
                have to correspond to the dimension names.
            extract_metadata (bool, optional): If True, missing ``dims`` or
                ``coords`` arguments are tried to be populated from the
                container attributes.
            apply_metadata (bool, optional): Whether to apply the extracted
                or passed ``dims`` and ``coords`` to the underlying data.
                This might not be desired in cases where the given ``data``
                already is a labelled ``xr.DataArray`` or where the data is a
                proxy and the labelling should be postponed.
            **dc_kwargs: passed to parent
        """
        # To be a bit more tolerant, allow lists as data argument
        if isinstance(data, list):
            log.debug("Received a list as `data` argument to %s '%s'. "
                      "Calling np.array on it ...", self.classname, name)
            data = np.array(data)

        # Initialize with parent method
        super().__init__(name=name, data=data, **dc_kwargs)
        # NOTE The _data attribute is now set, but will be changed again below!

        # Set up cache attributes with given arguments
        self._dim_names = dims
        self._dim_to_coords_map = coords

        # Keep track of whether metadata was applied or not
        self._metadata_was_applied = False

        # If metadata is to be extracted from container attributes, do so now
        if extract_metadata:
            self._extract_metadata()

        # Apply the metadata, if set to do so (and not a proxy, which would not
        # allow it) ...
        if apply_metadata and not isinstance(self._data, AbstractDataProxy):
            self._apply_metadata()

    def _format_info(self) -> str:
        """A __format__ helper function: returns info about the item.

        In this case, the dtype and sizes of the stored data is returned.
        Depending on whether metadata is available, the shape information is
        shown or the dimension names and the length of the dimensions are used.
        """
        return ("{dtype:}, {shape:}, {sup:}"
                "".format(dtype=self.dtype, shape=self._format_shape(),
                          sup=super()._format_info()))

    def _format_shape(self) -> str:
        """A __format__ helper for parsing shape information"""
        sizes = None
        if self._metadata_was_applied:
            # Can directly use the xarray information
            sizes = self.sizes.items()

        elif self._dim_names is not None:
            # Parse the metadata accordingly ...
            sizes = self._parse_sizes_from_metadata()

        if sizes is not None:
            return "({})".format(", ".join(["{}: {}".format(*kv)
                                            for kv in sizes]))
        return "shape {}".format(self.shape)

    def copy(self):
        """Return a new object with a deep copy of the data.
        
        Returns:
            XrDataContainer: A deep copy of this object.
        """
        log.debug("Creating copy of %s ...", self.logstr)

        return self.__class__(name=self.name + "_copy",
                              data=copy.deepcopy(self._data),
                              attrs=copy.deepcopy(self.attrs),
                              # Carry over cache attributes, needed for proxy
                              dims=self._dim_names,
                              coords=self._dim_to_coords_map,
                              # No need to extract or apply; if it is a proxy,
                              # the metadata is passed; if it is an xarray, it
                              # is already labelled and passed via `data`
                              extract_metadata=False,
                              apply_metadata=False)

    def save(self, path: str, **save_kwargs):
        """Saves the XrDataContainer to a file by invoking the .to_netcdf
        method of the underlying data.
        
        The recommended file extension is ``.xrdc`` or ``.nc_da``, which are
        compatible with the xarray-based data loader.
        
        .. warning::
        
            This does NOT store container attributes!
        
        Args:
            path (str): The path to save the file at
            **save_kwargs: Passed to the .to_netcdf method
        """
        self.to_netcdf(path, **save_kwargs)

    # Methods to extract and apply metadata ...................................

    def _extract_metadata(self):
        """Extracts metadata from the container attributes and stores them
        in the ``_dim_names`` and ``_dim_to_coords_map`` cache attributes.
        """
        log.trace("Extracting metadata for labelling %s ...", self.logstr)
        
        # First: the dimension names
        if self._dim_names is None:
            try:
                dims = extract_dim_names(self.attrs, ndim=self.ndim,
                                         attr_name=self._XRC_DIMS_ATTR,
                                         attr_prefix=self._XRC_DIM_NAME_PREFIX)

            except Exception as exc:
                raise type(exc)("Failed extracting dimension names from the "
                                "attributes of {}! {}"
                                "".format(self.logstr, exc)) from exc
            
            else:
                self._dim_names = dims

        # With dimension names being cached, try extracting coordinates.
        if self._dim_to_coords_map is None:
            coords = extract_coords(self, mode='attrs',
                        dims=self._dim_names,
                        # Attribute names and prefixes
                        strict=self._XRC_STRICT_ATTR_CHECKING,
                        coords_attr_prefix=self._XRC_COORDS_ATTR_PREFIX,
                        mode_attr_prefix=self._XRC_COORDS_MODE_ATTR_PREFIX,
                        default_mode=self._XRC_COORDS_MODE_DEFAULT)
            self._dim_to_coords_map = coords

    def _inherit_attrs(self):
        """Carry over container attributes to the xr.DataArray attributes

        This does not include container attributes that are used for extracting
        metadata; it makes no sense to have them in the attributes of the
        already labelled xr.DataArray
        """
        def skip(attr_name: str) -> bool:
            return (   attr_name == self._XRC_DIMS_ATTR
                    or attr_name.startswith(self._XRC_DIM_NAME_PREFIX)
                    or attr_name.startswith(self._XRC_COORDS_ATTR_PREFIX)
                    or attr_name.startswith(self._XRC_COORDS_MODE_ATTR_PREFIX))

        for attr_name, attr_val in self.attrs.items():
            if not skip(attr_name):
                self.data.attrs[attr_name] = attr_val
    
    def _apply_metadata(self):
        """Applies the cached metadata to the underlying xr.DataArray"""
        # Make sure that data is an xarray
        if not isinstance(self.data, xr.DataArray):
            self._data = xr.DataArray(self.data)

        # Carry over the name
        self.data.name = self.name

        # Set the dimension names
        if self._dim_names:
            # Create a mapping from old to new names, then apply it
            new_names = {old: new
                         for old, new in zip(self.data.dims, self._dim_names)
                         if new is not None}
            
            log.trace("Renaming dimensions:  %s", new_names)
            self._data = self.data.rename(new_names)

        # Set the coordinates
        if self._dim_to_coords_map:
            log.trace("Associating coordinates:  %s", self._dim_to_coords_map)

            for dim_name, coords in self._dim_to_coords_map.items():
                # Need to handle links differently
                if isinstance(coords, Link):
                    # The target object is another DataContainer, which can not
                    # be used for association. Thus, just pass the raw data...
                    coords = coords.target_object.values  # np.ndarray now

                # Can associate now.
                try:
                    self.data.coords[dim_name] = coords
                
                except Exception as err:
                    raise ValueError("Could not associate coordinates {} for "
                                     "dimension '{}' due to a {}: {}."
                                     "".format(coords, dim_name,
                                               err.__class__.__name__, err)
                                     ) from err
        
        # Now write the rest of the attributes of the dataset to the xarray
        if self._XRC_INHERIT_CONTAINER_ATTRIBUTES:
            self._inherit_attrs()

        # Now set the flag that metadata was applied
        self._metadata_was_applied = True

    def _postprocess_proxy_resolution(self):
        """Only invoked from
        :py:class:`~dantro.mixins.proxy_support.ProxySupportMixin`, which have
        to be added to the class specifically. This function takes care to
        apply the potentially existing metadata *after* the proxy was resolved.
        """
        self._apply_metadata()

    def _parse_sizes_from_metadata(self) -> Sequence[Tuple[str, int]]:
        """Invoked from _format_shape when no metadata was applied but the
        dimension names are available. Should return data in the same form as
        xr.DataArray.sizes.items() does.
        """
        # Iterate over dimension names and shapes ...
        it = enumerate(zip(self._dim_names, self.shape))

        # ... and use the name from the metadata unless the name was None
        # which is a placeholder for "don't rename this dimension", in which
        # case it should be named via default names
        return tuple([(n if n else "dim_{}".format(i), l) for i, (n, l) in it])



"""This module implements the PlotManager class, which handles the
configuration of multiple plots and prepares the data and configuration to pass
to the PlotCreator.
"""

import os
import time
import copy
import logging
from typing import Union, List, Dict, Tuple, Callable, Any

from paramspace import ParamSpace, ParamDim

from .data_mngr import DataManager
from .plot_creators import ALL as ALL_PCRS
from .plot_creators import BasePlotCreator
from .tools import load_yml, write_yml, recursive_update

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Custom exception classes

class PlottingError(Exception):
    """Custom exception class for all plotting errors"""

class PlotConfigError(ValueError, PlottingError):
    """Raised when there were errors in the plot configuration"""

class InvalidCreator(ValueError, PlottingError):
    """Raised when an invalid creator was specified"""

class PlotCreatorError(PlottingError):
    """Raised when an error occured in a plot creator"""

# -----------------------------------------------------------------------------

class PlotManager:
    """The PlotManager takes care of configuring plots and calling the
    configured PlotCreator classes that then carry out the plots.
    
    Attributes:
        CREATORS (dict): The mapping of creator names to classes. When it is
            desired to subclass PlotManager and extend the creator mapping, use
            `dict(**pcr.ALL)` to inherit the default creator mapping.
        DEFAULT_OUT_FSTRS (dict): The default values for the output format
            strings.
    """

    CREATORS = ALL_PCRS
    DEFAULT_OUT_FSTRS = dict(timestamp="%y%m%d-%H%M%S",
                             state_no="{no:0{digits:d}d}",
                             state="{name:}_{val:}",
                             state_name_replace_chars=[], # (".", "-")
                             state_val_replace_chars=[("/", "-")],
                             state_join_char="__",
                             state_vector_join_char="-",
                             # final fstr for single plot and config path
                             path="{name:}{ext:}",
                             plot_cfg="{basename:}_cfg.yml",
                             # and for sweep
                             sweep="{name:}/{state_no:}__{state:}{ext:}",
                             plot_cfg_sweep="{name:}/sweep_cfg.yml",
                             )

    def __init__(self, *, dm: DataManager, base_cfg: Union[dict, str]=None,
                 update_base_cfg: Union[dict, str]=None,
                 plots_cfg: Union[dict, str]=None,
                 out_dir: Union[str, None]="{timestamp:}/",
                 out_fstrs: dict=None,
                 creator_init_kwargs: Dict[str, dict]=None,
                 default_creator: str=None,
                 auto_detect_creator: bool=False,
                 save_plot_cfg: bool=True,
                 raise_exc: bool=False,
                 cfg_exists_action: str='raise'):
        """Initialize the PlotManager

        The initialization comes with three (optional) hierarchical levels to
        make the configuration of plots versatile, flexible, and avoid copy-
        paste of configurations:
        The first two result in a so-called "base" configuration, a collection
        of available, but disabled plot configs.
        The third specifies the default plots, which can use the `based_on`
        feature to base their configuration on any of the configurations from
        the base plot configuration.

        Specifically:

            1. The ``base_cfg`` contains a set of plot configurations that form
                a repertoire of configurations. These are not performed by
                default, but can be imported.
            2. The ``update_base_cfg`` contains plot configurations that are
                possibly derived from the base repertoire. This happens in the
                following way: First by recursive update of existing entries,
                and second by resolving the ``based_on: a_base_plot`` again by
                recursive update.
            3. The ``plots_cfg`` holds enabled plot configurations, possibly
                derived from the base configuration using the `based_on`
                feature, e.g. ``based_on: a_base_plot``; this happens by
                recursive update.

        Args:
            dm (DataManager): The DataManager-derived object to read the plot
                data from.
            base_cfg (Union[dict, str], optional): The default base config or a
                path to a yaml-file to import. The base config defines a set
                of plot configuration that other plot configurations can
                declare themselves ``based_on``.
            update_base_cfg (Union[dict, str], optional): An update config to
                the base config or a path to a yaml-file to import which
                recursively updates the ``base_cfg``.
            plots_cfg (Union[dict, str], optional): The default plots config or
                a path to a yaml-file to import
            out_dir (Union[str, None], optional): If given, will use this
                output directory as basis for the output path for each plot.
                The path can be a format-string; it is evaluated upon call to
                the plot command. Available keys: ``timestamp``, ``name``, ...
                For a relative path, this will be relative to the DataManager's
                output directory. Absolute paths remain absolute.
                If this argument evaluates to False, the DataManager's output
                directory will be the output directory.
            out_fstrs (dict, optional): Format strings that define how the
                output path is generated. The dict given here updates the
                ``DEFAULT_OUT_FSTRS`` class variable which holds the default
                values.

                Keys: ``timestamp`` (%-style), ``path``, ``sweep``, ``state``,
                ``plot_cfg``, ``state``, ``state_no``, ``state_join_char``,
                ``state_vector_join_char``.

                Available keys for ``path``: ``name``, ``timestamp``, ``ext``.

                Additionally, for ``sweep``: ``state_no``, ``state_vector``,
                    ``state``.

            creator_init_kwargs (Dict[str, dict], optional): If given, these
                kwargs are passed to the initialization calls of the respective
                creator classes.
            default_creator (str, optional): If given, a plot without explicit
                ``creator`` declaration will use this creator as default.
            auto_detect_creator (bool, optional): If true, and no default
                creator is given, will try to automatically deduce the creator
                using the given plot arguments. All creators registered with
                this PlotManager instance are candidates.
            save_plot_cfg (bool, optional): If True, the plot configuration is
                saved to a yaml file alongside the created plot.
            raise_exc (bool, optional): Whether to raise exceptions if there
                are errors raised from the plot creator or errors in the plot
                configuration. If False, the errors will only be logged.
            cfg_exists_action (str, optional): Behaviour when a config file
                already exists. Can be: ``raise`` (default), ``skip``,
                ``append``, ``overwrite``, or ``overwrite_nowarn``.
        
        Raises:
            InvalidCreator: When an invalid default creator was chosen
            KeyError: Upon bad ``based_on`` in ``update_base_cfg``
        """
        # TODO consider making it possible to pass classes for plot creators

        # Initialize attributes and store arguments
        self._plot_info = []

        # Public
        self.save_plot_cfg = save_plot_cfg
        self.raise_exc = raise_exc

        # Private or read-only
        self._dm = dm
        self._out_dir = out_dir
        self._auto_detect_creator = auto_detect_creator

        # Parameters to pass through as defaults to member functions
        self._cfg_exists_action = cfg_exists_action

        # Handle base config
        if isinstance(base_cfg, str):
            # Interpret as path to yaml file
            log.debug("Loading base_cfg from file %s ...", base_cfg)
            self._base_cfg = load_yml(base_cfg)
        else:
            self._base_cfg = copy.deepcopy(base_cfg)

        # Handle the update of base config
        if isinstance(update_base_cfg, str):
            # Interpret as path to yaml file
            log.debug("Loading update_base_cfg from file %s ...",
                      update_base_cfg)
            update_base_cfg = load_yml(update_base_cfg)

        # Perform update of base config: recursive + resolution of `based_on`
        if update_base_cfg:
            # First, make a recursive update of the existing based_on
            self._base_cfg = recursive_update(self._base_cfg, update_base_cfg)
            # Now, potentially existing `based_on` entries from either of
            # these configurations are part of the _base_cfg
            # Resolve these `based_on` keys ...
            for pcfg_name, pcfg in self._base_cfg.items():
                based_on = pcfg.pop('based_on', None)

                if not based_on:
                    continue

                elif isinstance(based_on, str):
                    based_on = (based_on,)
                
                # Now a sequence of strings; go over all these and build a dict
                # that will be the base that is to be recursively updated with
                # the given update config.
                bcfg = dict()
                for _based_on in based_on:
                    if _based_on not in self._base_cfg:
                        raise KeyError("No base plot configuration named '{}' "
                                       "available to use during resolution of "
                                       "`update_base_cfg`! Available: {}"
                                       "".format(_based_on,
                                                 ", ".join(self._base_cfg)))

                    # Need to work on a deep copy of the original base config
                    # entry in order to not get any mutability issues
                    _bcfg = copy.deepcopy(self._base_cfg[_based_on])
                    bcfg = recursive_update(bcfg, _bcfg)
                
                # Finally, apply the given update and store as attribute
                self._base_cfg[pcfg_name] = recursive_update(bcfg, pcfg)

        # Handle default plots configuration
        if isinstance(plots_cfg, str):
            # Interpret as path to yaml file
            log.debug("Loading plots_cfg from file %s ...", plots_cfg)
            plots_cfg = load_yml(plots_cfg)
        self._plots_cfg = plots_cfg

        # Update the default format strings, if any were given here
        self._out_fstrs = self.DEFAULT_OUT_FSTRS
        if out_fstrs:
            self._out_fstrs = recursive_update(copy.deepcopy(self._out_fstrs),
                                               out_fstrs)

        # Store creator init kwargs
        self._cckwargs = creator_init_kwargs if creator_init_kwargs else {}

        # Store default creator name
        if default_creator and default_creator not in self.CREATORS:
            raise InvalidCreator("No such creator '{}' available, only: {}"
                                 "".format(default_creator,
                                           [k for k in self.CREATORS.keys()]))
        self._default_creator = default_creator

        log.debug("%s initialized.", self.__class__.__name__)

    # .........................................................................
    # Properties

    @property
    def out_fstrs(self) -> dict:
        """Returns the dict of output format strings"""
        return self._out_fstrs

    @property
    def plot_info(self) -> List[dict]:
        """Returns a list of dicts with info on all plots"""
        return self._plot_info

    @property
    def base_cfg(self) -> dict:
        """Returns a deep copy of the base configuration"""
        return copy.deepcopy(self._base_cfg)

    # .........................................................................
    # Helpers

    def _parse_out_dir(self, fstr: str, *, name: str) -> str:
        """Evaluates the format string to create an output directory.
        
        Note that the directories are _not_ created; this is outsourced to the
        plot creator such that it happens as late as possible.
        
        Args:
            fstr (str): The format string to evaluate and create a directory at
            name (str): Name of the plot
            timestamp (float, optional): Description
        
        Returns:
            str: The path of the created directory
        """
        # Get date format string and current time and create a string
        # TODO allow passing a timestamp?
        timefstr = self._out_fstrs.get('timestamp', "%y%m%d-%H%M%S")
        timestr = time.strftime(timefstr)

        out_dir = fstr.format(timestamp=timestr, name=name)

        # Make sure it is absolute
        out_dir = os.path.expanduser(out_dir)
        if not os.path.isabs(out_dir):
            # Regard it as relative to the data manager's output directory
            out_dir = os.path.join(self._dm.dirs['out'], out_dir)

        # Return the full path
        return out_dir

    def _parse_out_path(self, creator: BasePlotCreator, *, name: str,
                        out_dir: str, file_ext: str=None, state_no: int=None,
                        state_no_max: int=None, state_vector: Tuple[int]=None,
                        dims: dict=None) -> str:
        """Given a creator and (optionally) parameter sweep information, a full
        and absolute output path is generated, including the file extension.
        
        Note that the directories are _not_ created; this is outsourced to the
        plot creator such that it happens as late as possible.
        
        Args:
            creator (BasePlotCreator): The creator instance, used to
                extract information on the file extension.
            name (str): The name of the plot
            out_dir (str): The absolute output directory, prepended to all
                generated paths
            file_ext (str, optional): The file extension to use
            state_no (int, optional): The state number, starting with 0
            state_no_max (int, optional): The maximum state number
            state_vector (Tuple[int], optional): The state vector with info
                on how far each state dimension has progressed in the sweep
            dims (dict, optional): The dict of parameter dimensions of the
                sweep that is carried out.
        
        Returns:
            str: The fully parsed output path for this plot
        """
        def parse_state_pair(name: str, dim: ParamDim, *,
                             fstrs: dict) -> Tuple[str]:
            """Helper method to create a state pair"""
            # Parse the name
            for search, replace in fstrs['state_name_replace_chars']:
                name = name.replace(search, replace)

            # Parse the value
            val = str(dim.current_value)
            for search, replace in fstrs['state_val_replace_chars']:
                val = val.replace(search, replace)
            
            return name, val


        # Get the fstrs
        fstrs = self.out_fstrs
        
        # Evaluate the keys available for both cases
        keys = dict(timestamp=time.strftime(fstrs['timestamp']), name=name)

        # Parse file extension and ensure it starts with a dot
        ext = file_ext if file_ext else creator.get_ext()

        if ext and ext[0] != ".":
            ext = "." + ext
        elif ext is None:
            ext = ""

        keys['ext'] = ext


        # Change behaviour depending on whether state information was given
        if state_no is None:
            # Assume the other arguments are also None -> Not part of the sweep
            # Evaluate it
            out_path = fstrs['path'].format(**keys)

        else:
            # Is part of a sweep
            # Parse additional keys
            # state number
            digits = len(str(state_no_max))
            keys['state_no'] = fstrs['state_no'].format(no=state_no,
                                                        digits=digits)

            # state values -- need to do some parsing here ...
            state_pairs = [parse_state_pair(name, dim, fstrs=fstrs)
                           for name, dim in dims.items()]

            sjc = fstrs['state_join_char']
            keys['state'] = sjc.join([fstrs['state'].format(name=k, val=v)
                                      for k, v in state_pairs])

            # state vector
            svjc = fstrs['state_vector_join_char']
            keys['state_vector'] = svjc.join([str(s) for s in state_vector])

            # Evaluate it
            out_path = fstrs['sweep'].format(**keys)

        # Prepend the output directory and return
        out_path = os.path.join(out_dir, out_path)

        return out_path
    
    def _resolve_based_on(self, *, cfg: dict,
                          based_on: Union[str, Tuple[str]]=None,
                          work_on_deepcopy: bool=True) -> dict:
        """Resolves the ``based_on`` reference in a plot configuration
        
        Args:
            cfg (dict): The plot configuration that will be used to recursively
                update the specified base configurations.
            based_on (Union[str, Tuple[str]], optional): The name or names of
                the base configuration entries to use for updating
            work_on_deepcopy (bool, optional): Whether to work on a deepcopy
        
        Returns:
            dict: The plot configuration derived from the one at ``based_on``
        
        Raises:
            KeyError: If ``based_on`` value is not a key in ``self._base_cfg``
        """
        if not based_on:
            return cfg

        elif isinstance(based_on, str):
            based_on = (based_on,)

        # Copy, if needed
        if work_on_deepcopy:
            cfg = copy.deepcopy(cfg)

        # Resolve the list of based_on entries
        base_cfg = dict()
        for _based_on in based_on:
            if _based_on not in self.base_cfg.keys():
                raise KeyError("No base plot configuration named '{}' "
                               "available! Choose from: {}"
                               "".format(_based_on,
                                         ", ".join(self._base_cfg.keys())))

            # Do the recursive update. The base_cfg property already returns
            # a deep copy of the base configuration ...
            base_cfg = recursive_update(base_cfg, self.base_cfg[_based_on])

        # As final step, update the base configuration with everything 
        return recursive_update(base_cfg, cfg)

    def _get_plot_creator(self, creator: Union[str, None],
                          *, name: str, init_kwargs: dict,
                          from_pspace: ParamSpace=None, plot_cfg: dict,
                          auto_detect: bool=None) -> BasePlotCreator:
        """Determines which plot creator to use by looking at the given
        arguments. If set, tries to auto-detect from the arguments, which
        creator is to be used.
        
        Then, sets up the corresponding creator and returns it.

        This method is called from the plot() method.
        
        Args:
            creator (Union[str, None]): The name of the creator to be found.
                Can be None, if no argument was given to the plot method.
            name (str): The name of the plot
            init_kwargs (dict): Additional creator initialization parameters
            from_pspace (ParamSpace, optional): If the plot is to be creatd
                from a parameter space, that parameter space.
            plot_cfg (dict): The plot configuration
            auto_detect (bool, optional): Whether to auto-detect the creator.
                If none, the value given at initialization is used.
        
        Returns:
            BasePlotCreator: The selected creator object, fully initialized.
        
        Raises:
            InvalidCreator: If the ``creator`` argument was invalid or auto-
                detection failed.
        """
        # If no creator is given, check if a default one can be used
        if creator is None:
            # Combine auto-detect related settings
            auto_detect = (auto_detect if auto_detect is not None
                           else self._auto_detect_creator)

            # Find other ways to determine the name of the creator
            if self._default_creator:
                # Just use the default
                creator = self._default_creator

            elif auto_detect:
                # Can try to auto-detect from the arguments, which creator
                # could uniquely fit to the arguments
                log.debug("Attempting auto-detection of creator ...")

                # If a ParamSpace plot is to be made, detect feasibility by
                # using its default parameters
                cfg = plot_cfg if not from_pspace else from_pspace.default

                # (name, plot creator) tuples for each candidate
                pc_candidates = []
                
                # Go over all registered plot creators
                for pc_name in self.CREATORS.keys():
                    try:
                        # Instantiate them by calling this function recursively
                        pc = self._get_plot_creator(pc_name, name=name,
                                                    init_kwargs=init_kwargs,
                                                    from_pspace=from_pspace,
                                                    plot_cfg=plot_cfg)
                        # NOTE Cannot call a class method here, because some
                        #      creators might need the actual arguments in
                        #      order to work properly

                    except:
                        # Failed to initialize for whatever reason, thus not
                        # a candidate
                        continue

                    # Successfully initialized. Check if it's a candidate for
                    # this plot configuration
                    if pc.can_plot(pc_name, **cfg):
                        log.debug("Plot creator '%s' declared itself a "
                                  "candidate.", pc_name)
                        pc_candidates.append((pc_name, pc))

                # If there is more than one candidate, cannot decide
                if len(pc_candidates) > 1:
                    pcc_names = [n for n, _ in pc_candidates]
                    raise InvalidCreator("Tried to auto-detect a plot creator "
                                         "for plot '{}' but could not "
                                         "unambiguously do so! There were {} "
                                         "plot creators declaring themselves "
                                         "as candidates: {}. Consider "
                                         "specifying the creator explicitly."
                                         "".format(name, len(pc_candidates),
                                                   ", ".join(pcc_names)))
                elif len(pc_candidates) < 1:
                    pc_names = ", ".join([k for k in self.CREATORS.keys()])
                    raise InvalidCreator("Tried to auto-detect a plot creator "
                                         "for plot '{}' but none of the "
                                         "available creators ({}) declared "
                                         "itself a candidate! This might also "
                                         "be due to a plot configuration that"
                                         "the candidate plot creators could "
                                         "not interpret. Consider specifying "
                                         "the creator explicitly to find out "
                                         "why auto-detection failed."
                                         "".format(name, pc_names))

                # else: there was only one, use that
                pc_name, pc = pc_candidates[0]
                log.debug("Auto-detected plot creator: %s", pc.logstr)

                # As it is already initialized, can just return it
                return pc

            else:
                raise InvalidCreator("No `creator` argument given and neither "
                                     "`default_creator` specified during "
                                     "initialization nor auto-detection "
                                     "enabled. Cannot plot!")

        # Parse initialization kwargs, based on the defaults set in __init__
        pc_kwargs = self._cckwargs.get(creator, {})
        if init_kwargs:
            log.debug("Recursively updating creator initialization kwargs ...")
            pc_kwargs = recursive_update(copy.deepcopy(pc_kwargs), init_kwargs)
        
        # Instantiate the creator class
        pc = self.CREATORS[creator](name=name, dm=self._dm, **pc_kwargs)

        log.debug("Initialized %s.", pc.logstr)
        return pc

    def _invoke_creator(self, plot_creator: Callable, *, out_path: str,
                        **plot_cfg) -> Any:
        """This method wraps the plot creator's ``__call__`` and is the last
        PlotManager method that is called prior to handing over to the selected
        plot creator. It takes care of invoking the plot creator's ``__call__``
        method and handling potential error messages and return values.
        
        Args:
            plot_creator (Callable): The currently used creator
            out_path (str): The plot output path
            **plot_cfg: The plot configuration
        
        Returns:
            Any: The return value of the plot creator's ``__call__`` method
        
        Raises:
            PlotCreatorError: On error within the plot creator
        """
        try:
            rv = plot_creator(out_path=out_path, **plot_cfg)

        except Exception as err:
            # No return value
            rv = None

            # Generate error message
            e_msg = ("During plotting with {}, a {} occurred: {}"
                     "".format(plot_creator.logstr,
                               err.__class__.__name__, str(err)))

            if self.raise_exc:
                raise PlotCreatorError(e_msg) from err
            
            # else: just log it
            log.error(e_msg)

        else:
            log.debug("Plot creator call returned.")

        return rv

    def _store_plot_info(self, name: str,
                         *, plot_cfg: dict, creator_name: str, save: bool,
                         target_dir: str, **info):
        """Stores all plot information in the plot_info list and, if `save` is
        set, also saves it using the _save_plot_cfg method.
        """
        # Prepare the entry
        entry = dict(name=name, plot_cfg=plot_cfg, target_dir=target_dir,
                     creator_name=creator_name, **info,
                     plot_cfg_path=None)

        if save:
            # Save the plot configuration
            save_path = self._save_plot_cfg(plot_cfg, name=name,
                                            target_dir=target_dir,
                                            creator_name=creator_name)

            # Store the path the configuration was saved at
            entry['plot_cfg_path'] = save_path

        # Append to the plot_info list
        self._plot_info.append(entry)

    def _save_plot_cfg(self, cfg: dict,
                       *, name: str, creator_name: str, target_dir: str,
                       exists_action: str=None, is_sweep: bool=False) -> str:
        """Saves the given configuration under the top-level entry ``name`` to
        a yaml file.
        
        Args:
            cfg (dict): The plot configuration to save
            name (str): The name of the plot
            creator_name (str): The name of the creator
            target_dir (str): The directory path to store the file in
            exists_action (str, optional): What to do if a plot configuration
                already exists. Can be: ``overwrite``, ``overwrite_nowarn``,
                ``skip``, ``append``, ``raise``. If None, uses the value of the
                ``cfg_exists_action`` argument given during initialization.
            is_sweep (bool, optional): Set if the configuration refers to a
                plot in sweep mode, for which a different format string is used
        
        Returns:
            str: The path the config was saved at (mainly used for testing)
        
        Raises:
            ValueError: For invalid ``exists_action`` argument
        """
        # Resolve default arguments
        if exists_action is None:
            exists_action = self._cfg_exists_action

        # Build the dict that is to be saved
        d = dict()
        d[name] = copy.deepcopy(cfg)

        if not isinstance(cfg, ParamSpace):
            d[name]['creator'] = creator_name

        else:
            # FIXME hacky, should not use the internal API!
            d[name]._dict['creator'] = creator_name

        # Generate the filename and save path
        fn_fstr = self.out_fstrs['plot_cfg_sweep' if is_sweep else 'plot_cfg']
        fname = fn_fstr.format(name=name,
                               basename=os.path.basename(name))
        save_path = os.path.join(target_dir, fname)
        
        # Try to write
        try:
            write_yml(d, path=save_path, mode='x')

        except FileExistsError as err:
            log.debug("Config file already exists at %s!", save_path)
            
            if exists_action == 'raise':
                raise

            elif exists_action == 'skip':
                log.debug("Skipping ...")

            elif exists_action == 'append':
                log.debug("Appending ...")
                write_yml(d, path=save_path, mode='a')

            elif exists_action == 'overwrite':
                log.warning("Overwriting existing plot configuration ...")
                write_yml(d, path=save_path, mode='w')

            elif exists_action == 'overwrite_nowarn':
                log.debug("Overwriting ...")
                write_yml(d, path=save_path, mode='w')

            else:
                raise ValueError("Invalid value '{}' for argument "
                                 "`exists_action`!".format(exists_action))

        else:
            log.debug("Saved plot configuration for '%s' to: %s",
                      name, save_path)

        return save_path


    # .........................................................................
    # Plotting

    def plot_from_cfg(self, *,
                      plots_cfg: Union[dict, str]=None,
                      plot_only: List[str]=None,
                      out_dir: str=None,
                      **update_plots_cfg) -> None:
        """Create multiple plots from a configuration, either a given one or
        the one passed during initialization.
        
        This is mostly a wrapper around the plot function, allowing additional
        ways of how to configure and create plots.
        
        Args:
            plots_cfg (dict, optional): The plots configuration to use. If not
                given, the one specified during initialization is used. If a
                string is given, will assume it is a path and load the file.
            plot_only (List[str], optional): If given, create only those plots
                from the resulting configuration that match these names. This
                will lead to the `enabled` key being ignored, regardless of its
                value.
            out_dir (str, optional): A different output directory; will use the
                one passed at initialization if the given argument evaluates to
                False.
            **update_plots_cfg: If given, it is used to update the plots_cfg
                recursively. Note that on the top level the _names_ of the
                plots are placed; this cannot be used to make all plots have a
                common property.
        
        Raises:
            PlotConfigError: Empty or invalid plot configuration
        """
        # Determine which plot configuration to use
        if not plots_cfg:
            if not self._plots_cfg and not update_plots_cfg:
                e_msg = ("Got empty `plots_cfg` and `plots_cfg` given at "
                         "initialization was also empty. Nothing to plot.")

                if self.raise_exc:
                    raise PlotConfigError(e_msg)

                log.error(e_msg)
                return

            log.debug("No new plots configuration given; will use plots "
                      "configuration given at initialization.")
            plots_cfg = self._plots_cfg

        elif isinstance(plots_cfg, str):
            # Interpret as path to yaml file
            log.debug("Loading plots_cfg from file %s ...", plots_cfg)
            plots_cfg = load_yml(plots_cfg)

        # Make sure to work on a copy, be it on the defaults or on the passed
        plots_cfg = copy.deepcopy(plots_cfg)

        if update_plots_cfg:
            # Recursively update with the given keywords
            plots_cfg = recursive_update(plots_cfg, update_plots_cfg)
            log.debug("Updated the plots configuration.")

        # Check the plot configuration for invalid types
        for plot_name, cfg in plots_cfg.items():
            if not isinstance(cfg, (dict, ParamSpace)):
                raise PlotConfigError("Got invalid plots specifications for "
                                      "entry '{}'! Expected dict, got {} with "
                                      "value '{}'. Check the correctness of "
                                      "the given plots configuration!"
                                      "".format(plot_name, type(cfg), cfg))

        # Filter the plot selection
        if plot_only is not None:
            # Only plot these entries
            plots_cfg = {k:plots_cfg[k] for k in plot_only}
            # NOTE that this deliberately raises an error for an invalid entry
            #      in the `plot_only` argument

            # Remove all `enabled` keys from the remaining entries
            for cfg in plots_cfg.values():
                cfg.pop('enabled', None)

        else:
            # Resolve all `enabled` entries, creating a new plots_cfg dict
            plots_cfg = {k:v for k, v in plots_cfg.items()
                         if v.pop('enabled', True)}

        # Throw out entries that start with an underscore
        plots_cfg = {k:v for k, v in plots_cfg.items()
                     if not k.startswith("_")}

        # Determine and create the plot directory to use
        if not out_dir:
            out_dir = self._out_dir
        out_dir = self._parse_out_dir(out_dir, name="{name:}")
        # NOTE creating this here such that all plots from this config are side
        #      by side in one output directory. With the given `name` key, the
        #      evaluation of that part of the out_dir path is postponed
        #      to when the actual plot with that name is created.

        log.hilight("Performing plots from %d plot configuration entr%s ...",
                    len(plots_cfg), "ies" if len(plots_cfg) != 1 else "y")

        # Loop over the configured plots
        for plot_name, cfg in plots_cfg.items():
            # Use the public methods to perform the plotting call, depending
            # on the type of the config
            if isinstance(cfg, ParamSpace):
                # Is a parameter space. Use the corresponding call signature
                self.plot(plot_name, out_dir=out_dir, from_pspace=cfg)
            
            else:
                # Just a dict. Use the regular call
                self.plot(plot_name, out_dir=out_dir, **cfg)

        # All done
        log.success("Successfully performed plots for %d plot "
                    "configuration%s.\n",
                    len(plots_cfg), "s" if len(plots_cfg) != 1 else "")

    def plot(self, name: str, *,
             based_on: Union[str, Tuple[str]]=None,
             from_pspace: Union[dict, ParamSpace]=None,
             **plot_cfg) -> BasePlotCreator:
        """Create plot(s) from a single configuration entry.
        
        A call to this function resolves the `based_on` feature and passes the
        derived plot configuration to self._plot(), which actually carries out
        the plots.
        
        Note that more than one plot can result from a single configuration
        entry, e.g. when plots were configured that have more dimensions than
        representable in a single file.        
        
        Args:
            name (str): The name of this plot
            based_on (Union[str, Tuple[str]], optional): A key or a sequence
                of keys of entries in the base config that should be used as
                the basis of this plot.
                The given plot configuration is then used to recursively
                update (a copy of) those base configuration entries.
            from_pspace (Union[dict, ParamSpace], optional): If given, execute
                a parameter sweep over these parameters, re-using the same
                creator instance. If this is a dict, a ParamSpace is created
                from it.
            **plot_cfg: The plot configuration, including some parameters that
                the plot manager will evaluate (and consequently: does not
                pass on to the plot creator)
        
        Returns:
            BasePlotCreator: The PlotCreator used for these plots
        """
        # Derive the plot_cfg using based_on. Do this first only for the case
        # of the non-ParamSpace plot configuration, because it's easier.
        plot_cfg = self._resolve_based_on(cfg=plot_cfg, based_on=based_on)
        
        if from_pspace is None:
            # Can just invoke the helper and be done with it
            return self._plot(name, from_pspace=from_pspace, **plot_cfg)

        # Else: It's more complicated now, as the config is in from_pspace, and
        # (partly) in plot_cfg. Urgh.
        # Distinguish between dict and actual ParamSpace objects
        if isinstance(from_pspace, ParamSpace):
            # Already is a Paramspace. If so, will need to extract the
            # underlying dict to be able to do a recursive update
            pspace_plot_cfg = copy.deepcopy(from_pspace._dict)
            # FIXME Should not have to use private API!

            # Resolve `based_on`
            based_on = pspace_plot_cfg.pop("based_on", None)
            pspace_plot_cfg = self._resolve_based_on(cfg=pspace_plot_cfg,
                                                     based_on=based_on,
                                                     work_on_deepcopy=False)

            # Extract info, then re-create the ParamSpace with the updated cfg
            creator = pspace_plot_cfg.pop("creator", None)
            from_pspace = ParamSpace(pspace_plot_cfg)

        else:
            # Assume it's something dict-like
            # ... but need a copy in order to safely pop elements.
            from_pspace = copy.deepcopy(from_pspace)
            based_on = from_pspace.pop("based_on", None)

            # Do the update; no deepcopy needed in there because done here
            from_pspace = self._resolve_based_on(cfg=from_pspace,
                                                 based_on=based_on,
                                                 work_on_deepcopy=False)

            # Now also extract the creator; might have come in only by the
            # based_on resolution
            creator = from_pspace.pop("creator", None)

        # NOTE creator needs to be singled out above because _plot is not able
        #      to extract it from whatever `from_pspace` is.

        # Now have all the information extracted / removed from from_pspace to
        # be ready to call _plot. Finally.
        return self._plot(name, creator=creator, from_pspace=from_pspace,
                          **plot_cfg) # **plot_cfg is anything remaining ...

    def _plot(self, name: str, *, creator: str=None, out_dir: str=None,
              from_pspace: ParamSpace=None,
              file_ext: str=None, save_plot_cfg: bool=None,
              auto_detect_creator: bool=None, creator_init_kwargs: dict=None, 
              **plot_cfg) -> BasePlotCreator:
        """Create plot(s) from a single configuration entry.
        
        A call to this function creates a single PlotCreator, which is also
        returned after all plots are finished.
        
        Note that more than one plot can result from a single configuration
        entry, e.g. when plots were configured that have more dimensions than
        representable in a single file.
        
        Args:
            name (str): The name of this plot
            creator (str, optional): The name of the creator to use. Has to be
                part of the CREATORS class variable. If not given, the argument
                `default_creator` given at initialization will be used.
            out_dir (str, optional): If given, will use this directory as out
                directory. If not, will use the default value given at
                initialization.
            file_ext (str, optional): The file extension to use, including the
                leading dot!
            from_pspace (ParamSpace, optional): If given, execute a parameter
                sweep over these parameters, re-using the same creator instance
            save_plot_cfg (bool, optional): Whether to save the plot config.
                If not given, uses the default value from initialization.
            auto_detect_creator (bool, optional): Whether to attempt auto-
                detection of the ``creator`` argument. If given, this argument
                overwrites the value given at PlotManager initialization.
            creator_init_kwargs (dict, optional): Passed to the plot creator
                during initialization. Note that the arguments given at
                initialization of the PlotManager are updated by this.
            **plot_cfg: The plot configuration to pass on to the plot creator.
        
        Returns:
            BasePlotCreator: The PlotCreator used for these plots
        
        Raises:
            PlotConfigError: If no out directory was specified here or at
                initialization.
        """

        log.debug("Preparing plot '%s' ...", name)

        # Check that the output directory is given
        if not out_dir:
            if not self._out_dir:
                raise PlotConfigError("No `out_dir` specified here and at "
                                      "initialization; cannot perform plot.")

            out_dir = self._out_dir

        # Whether to save the plot config
        if save_plot_cfg is None:
            save_plot_cfg = self.save_plot_cfg

        # Get the plot creator, either by name or using auto-detect feature
        plot_creator = self._get_plot_creator(creator, name=name,
                                              init_kwargs=creator_init_kwargs,
                                              from_pspace=from_pspace, 
                                              plot_cfg=plot_cfg,
                                              auto_detect=auto_detect_creator)

        # Let the creator process arguments
        plot_cfg, from_pspace = plot_creator.prepare_cfg(plot_cfg=plot_cfg,
                                                         pspace=from_pspace)

        # Distinguish single calls and parameter sweeps
        if not from_pspace:
            log.progress("Performing '%s' plot ...", name)

            # Generate the output path
            out_dir = self._parse_out_dir(out_dir, name=name)
            out_path = self._parse_out_path(plot_creator, name=name,
                                            out_dir=out_dir, file_ext=file_ext)

            # Call the plot creator to perform the plot, using the private
            # method to perform exception handling
            self._invoke_creator(plot_creator, out_path=out_path, **plot_cfg)

            # Store plot information
            self._store_plot_info(name=name, creator_name=creator,
                                  out_path=out_path, plot_cfg=plot_cfg,
                                  save=save_plot_cfg,
                                  target_dir=os.path.dirname(out_path))
            
            log.progress("Finished '%s' plot.\n", name)

        else:
            # Is a parameter sweep over the plot configuration.
            # NOTE The parameter space is allowed to have volume 0!

            # If it is not already a ParamSpace, create one; useful if not
            # calling from plot_from_cfg, but directly ...
            if not isinstance(from_pspace, ParamSpace):
                from_pspace = ParamSpace(from_pspace)

            # Extract some info
            psp_vol = from_pspace.volume
            psp_dims = from_pspace.dims

            # ... and provide it to the logger
            if psp_vol > 0:
                amap_coords = from_pspace.active_state_map.coords
                max_dname_len = max(len(n) for n in amap_coords.keys())

                log.progress("Performing %d '%s' plots ...", psp_vol, name)
                log.note("... iterating over parameter space:\n%s",
                         "\n".join(["  * {0:<{d:}} : {1:}"
                                    "".format(dim_name,
                                             ", ".join([str(c) for c in
                                                        coords.values]),
                                             d=max_dname_len)
                                    for dim_name, coords
                                    in amap_coords.items()]))

            else:
                log.progress("Performing '%s' plot ...", name)
                log.note("... from default point in zero-volume parameter "
                         "space.")

            # Parse the output directory, such that all plots are together in
            # one directory even if the timestamp varies
            out_dir = self._parse_out_dir(out_dir, name=name)

            # Create the iterator
            it = from_pspace.iterator(with_info=('state_no', 'state_vector',
                                                 'coords'))
            
            # ...and loop over all points:
            for n, (cfg, state_no, state_vector, coords) in enumerate(it):
                log.progress("Performing plot {n:{d:}d} / {v:} ..."
                             "".format(n=n+1, d=len(str(psp_vol)), v=psp_vol))
                log.note("Current coordinates:  %s",
                         ",  ".join("{}: {}".format(*kv)
                                    for kv in coords.items()))

                # Handle the file extension parameter; it might come from the
                # given configuration and then needs to be popped such that it
                # is not propagated to the plot creator.
                _file_ext = cfg.pop('file_ext', file_ext)

                # Generate the output path
                out_path = self._parse_out_path(plot_creator,
                                                name=name,
                                                out_dir=out_dir,
                                                file_ext=_file_ext,
                                                state_no=state_no,
                                                state_no_max=psp_vol-1,
                                                state_vector=state_vector,
                                                dims=psp_dims)

                # Call the plot creator to perform the plot, using the private
                # method to perform exception handling
                self._invoke_creator(plot_creator, out_path=out_path,
                                     **cfg, **plot_cfg)
                # NOTE The **plot_cfg is passed here in order to not loose any
                # arguments that might have been passed to it. While `cfg`
                # _should_ hold all the arguments from the parameter space
                # iteration, there might be more arguments in `plot_cfg`;
                # rather than disallowing this, we pass them on and forward
                # responsibility downstream ...

                # Store plot information
                self._store_plot_info(name=name, creator_name=creator,
                                      out_path=out_path, plot_cfg=plot_cfg,
                                      state_no=state_no,
                                      state_vector=state_vector,
                                      save=False, # TODO check if reasonable
                                      target_dir=os.path.dirname(out_path))

                # Done with these coordinates

            # Save the plot configuration alongside, if configured to do so
            if save_plot_cfg:
                self._save_plot_cfg(from_pspace, name=name,
                                    creator_name=creator,
                                    target_dir=out_dir, is_sweep=True)

            log.progress("Finished all '%s' plots.\n", name)

        # Done now. Return the plot creator object
        return plot_creator
